
from ..nlft import NonLinearFourierSequence
from ..poly import Polynomial

from .. import numerics as bd

def inlft(a: Polynomial, b: Polynomial):
    """Computes the Inverse Non-Linear Fourier Transform using the plain layer stripping algorithm.

    Args:
        a, b (Polynomial): The pair `(a, b)` is the NLFT we want to compute the sequence for.

    Note:
        This algorithm is guaranteed to be numerically stable only if `a` is outer.
        To generate an outer complementary polynomial, you can use `weiss.complete`.

    Returns:
        NonLinearFourierSequence: A sequence whose NLFT is equal to `(a, b)`.
    """
    if len(a.support()) != len(b.support()) or a.support().stop != 1:
        return ValueError("(a, b) must be in the image of the NLFT.")
    
    n = a.effective_degree()

    a_star = a.conjugate()
    b = b.duplicate()

    F = []
    for k in range(n+1):
        Fk = b[0]/a_star[0]
        F.append(Fk)

        s = bd.sqrt(1 + bd.abs2(Fk))
        a_star, b = (a_star + bd.conj(Fk) * b).truncate(0, n-k)/s, \
                    (b - Fk * a_star).truncate(b.support_start+1, b.support_start+n-k)/s

        b.support_start -= 1 # divide by z

    return NonLinearFourierSequence(F, support_start=b.support_start)



import itertools
from math import prod
from numbers import Number
from typing import Iterable

from . import numerics as bd
from .numerics.backend import generic_complex, generic_real

from .poly import Polynomial
from .util import flatten, next_power_of_two, sequence_shift

def minimal_covering_range(l):
    """Given a list of tuples `l` containing `range` objects, returns a tuple `t` containing `range`
    objects such that `t[k] is the minimal range containing `l[j][k]` for every `j`."""
    if len(l) == 0:
        return ()
    
    N = max(len(t) for t in l)

    return tuple(
        range(
            min((rng.start for rng in col if rng is not None), default=0),  
            max((rng.stop for rng in col if rng is not None), default=0)
        )
        for col in zip(*[t + (None,) * (N - len(t)) for t in l])
    )

def deep_truncate(l, lens):
    """Returns the same multidimensional list truncates to the given lengths in each axis.

    Note:
        It is assumed that l has dimensions >= lens.
    
    Args:
        lens (tuple[int])"""
    if len(lens) > 1:
        ilens = lens[1:]
        return [deep_truncate(l[k], ilens) for k in range(lens[0])]
    
    return l[:lens[0]]

def deep_horner_eval(l, z: tuple[generic_complex]):
    """Evaluates the multivariate polynomial with coefficients l using Horner's method."""
    if len(z) == 0:
        return l
    
    evals = [deep_horner_eval(il, z[1:]) for il in l]
    
    res = evals[-1]
    for k in reversed(range(len(evals) - 1)):
        res = res * z[0] + evals[k]

    return res

def deep_sequence_shift(l, s: tuple[int]):
    """Cyclically shifts the elements along each axis, where k-th axis is shifted by k."""
    if isinstance(l, list) and len(l) != 0:
        if isinstance(l[0], Number):
            return sequence_shift(l, s[0])
    
    s1 = s[1:]
    return sequence_shift([deep_sequence_shift(il, s1) for il in l], s[0])

def deep_sequence_skip(l, s: tuple[int]):
    """Takes only every `s[k]` element of the list in each axis, the equivalent of l[::s] in multiple dimensions."""
    if isinstance(l, list) and len(l) != 0:
        if isinstance(l[0], Number):
            return l[::s[0]]
    
    s1 = s[1:]
    l = l[::s[0]]
    return [deep_sequence_skip(il, s1) for il in l]

def deep_inplace(l, func, reverse=False):
    """Applies the function to each element of the given nested list, in place."""
    if isinstance(l, list) and len(l) != 0:
        if reverse:
            l.reverse()

        if isinstance(l[0], Number):
            for k in range(len(l)):
                l[k] = func(l[k])
        else:
            for il in l:
                deep_inplace(il, func, reverse)

def deep_inplace_binary(l1, l2, func):
    """Applies the binary function to each element of the given nested lists, in place (they are assumed to be of the same dimension)."""
    if isinstance(l1, list) and len(l1) != 0:

        if isinstance(l1[0], Number):
            for k in range(len(l1)):
                l1[k] = func(l1[k], l2[k])
        else:
            for il1, il2 in zip(l1, l2):
                deep_inplace_binary(il1, il2, func)

def zeros(lens):
    if len(lens) == 1:
        return [bd.make_complex(0)] * (lens[0])
    
    zr = []
    for k in range(lens[0]):
        zr.append(zeros(lens[1:]))

    return zr

def sgn(k: int):
    if k > 0:
        return 1
    if k < 0:
        return -1
    return 0

def schwarz_multiplier(k: tuple):
    """Fourier multiplier for the Schwarz transform over several variables."""
    if all(c == 0 for c in k):
        return 1
    
    if all(c <= 0 for c in k):
        return 2
    
    return 0


class ComplexL0SequenceMD:

    def __init__(self, coeffs: list, support_start: tuple[int] | int):
        if isinstance(support_start, int):
            support_start = (support_start,)

        self.dim = len(support_start)
        self._xsupport_start = support_start[0]

        if not isinstance(coeffs, list):
            raise ValueError("Coefficient list must be of type list.")

        if self.dim == 1:
            if not all(isinstance(c, Number) for c in coeffs):
                raise ValueError("Coefficient list must be of the corresponding dimension.")
            
            self.coeffs = [bd.make_complex(c) for c in coeffs]
        else:
            self.coeffs = []

            for row in coeffs:
                if isinstance(row, list):
                    self.coeffs.append(self.__class__(row, support_start[1:]))
    
    def __xsupport(self):
        return range(self._xsupport_start, self._xsupport_start + len(self.coeffs))

    def support(self) -> tuple[range]:
        """Returns a tuple containing support ranges for all dimensions, i.e.,
        the hyper-parallelepiped in the grid containing all the coefficients of the polynomial."""
        xsupp = self.__xsupport()

        if self.dim == 1:
            return (xsupp,)
        
        return (xsupp, *minimal_covering_range([c.support() for c in self.coeffs]))
    
    @property
    def support_start(self):
        return tuple([t.start for t in self.support()])
    
    def coeff_list(self, rng=None):
        r"""Returns a `dim`-dimensional list containing the coefficients.

        Note:
            The stop element of the ranges are not included in the list.
        
        Args:
            rng: `dim`-dimensional tuple of range objects, giving the range for the hyper-parallelepiped along each axis. Defaults to the support of the sequence, as returned by `support()`.
        """
        if rng is None:
            rng = self.support()

        if self.dim == 1:
            return [self[k] for k in rng[0]]
        
        hcube = []
        lens = tuple(r.stop - r.start for r in rng[1:])
        for k in range(rng[0].start, rng[0].stop):

            if k - self._xsupport_start in range(len(self.coeffs)):
                hcube.append(self.coeffs[k - self._xsupport_start].coeff_list(rng[1:]))
            else:
                hcube.append(zeros(lens))
        
        return hcube
    
    def coeff_dict(self) -> dict:
        """Returns a dictionary of values representing the flattened version of `coeff_list()`.
        The keys are m-dimensional tuples, and the values are the corresponding coefficients."""
        L = self.coeff_list()
        result = {}

        def recurse(sublist, index_prefix=()):
            if not isinstance(sublist, list):
                # reached a coefficient value
                result[index_prefix] = sublist
                return
            for i, elem in enumerate(sublist):
                recurse(elem, index_prefix + (i,))

        recurse(L)
        return result
    
    def duplicate(self):
        return self.__class__(self.coeff_list(), self.support_start)
    
    def __getitem__(self, k: int) -> generic_complex:
        """Returns the coefficient in position (k1, ..., kd), or zero if the element is outside the support.
        """
        if not isinstance(k, tuple):
            k = (k,)

        if len(k) != self.dim:
            raise ValueError("Number of indices must coincide with the dimension of the sequence.")
        
        x = k[0]
        if x in self.__xsupport():
            if self.dim == 1:
                return self.coeffs[x - self._xsupport_start]
            
            return self.coeffs[x - self._xsupport_start][k[1:]]
        
        return bd.make_complex(0)

    def __setitem__(self, k: int, c: generic_complex):
        """Sets the coefficient of (k1, ..., kd) to be c, allocating space if needed.
        """
        x = k[0]
        if self.dim == 1:
            if self._xsupport_start + len(self.coeffs) <= x:
                self.coeffs.extend([bd.make_complex(0)] * (x - self._xsupport_start - len(self.coeffs) + 1))
            elif self._xsupport_start > x:
                self.coeffs = [bd.make_complex(0)] * (self._xsupport_start - x) + self.coeffs
                self._xsupport_start = x
            self.coeffs[x - self._xsupport_start] = bd.make_complex(c)
        else:
            if self._xsupport_start + len(self.coeffs) <= x:
                for _ in range(x - self._xsupport_start - len(self.coeffs) + 1):
                    self.coeffs.append(self.__class__([], support_start=(0,) * (self.dim-1)))
            elif self._xsupport_start > x:
                for _ in range(self._xsupport_start - x):
                    self.coeffs = [self.__class__([], support_start=(0,) * (self.dim-1))] + self.coeffs
                self._xsupport_start = x

            self.coeffs[x - self._xsupport_start][k[1:]] = bd.make_complex(c)

    def l1_norm(self) -> generic_real:
        """Computes the l1 norm of the sequence.

        Returns:
            float: The sum of absolute values of coefficients.
        """
        if self.dim == 1:
            return sum(bd.abs(c) for c in self.coeffs)
        return sum(c.l1_norm() for c in self.coeffs)

    def l2_norm(self) -> generic_real:
        """Computes the l2 norm.

        Returns:
            float: The l2 norm.
        """
        return bd.sqrt(self.l2_squared_norm())

    def l2_squared_norm(self) -> generic_real:
        """Computes the squared l2 norm.

        Returns:
            float: The squared l2 norm, i.e., the sum of the squared absolute values.
        """
        if self.dim == 1:
            return sum(bd.abs2(c) for c in self.coeffs)
        return sum(c.l2_squared_norm() for c in self.coeffs)
    
    def is_real(self) -> bool:
        """Whether the sequence has only real elements."""
        if self.dim == 1:
            return all(bd.im(F) <= bd.machine_threshold() for F in self.coeffs)
        return all(c.is_real() for c in self.coeffs)
    
    def is_imaginary(self) -> bool:
        """Whether the sequence has only imaginary elements."""
        if self.dim == 1:
            return all(bd.re(F) <= bd.machine_threshold() for F in self.coeffs)
        return all(c.is_imaginary() for c in self.coeffs)
    
    def _coeffwise_unary(self, func):
        """Returns a new sequence object `r` (as an object of the same class as self) whose coefficients are `r[k] = func(self[k])`.
        
        Note:
            It is implicitly assumed that func(0) == 0, so that compactness of the support is preserved."""
        cf = self.coeff_list()
        deep_inplace(cf, func)
        return self.__class__(cf, self.support_start)

    def _coeffwise_binary(self, other, func):
        """Returns a new sequence object `r` (as an object of the same class as self) whose coefficients are the pairwise `r[k] = func(self[k], other[k])`.
        
        Note:
            It is implicitly assumed that func(0, 0) == 0, i.e., the support of `r` will be the union of the supports of `p`, `q`."""
        union_support = tuple(range(min(x.start, y.start), max(x.stop, y.stop)) for x, y in zip(self.support(), other.support()))
        union_start = tuple(min(x, y) for x, y in zip(self.support_start, other.support_start))

        cf1 = self.coeff_list(union_support)
        cf2 = other.coeff_list(union_support)
        deep_inplace_binary(cf1, cf2, func)

        return self.__class__(cf1, union_start)
    
    def __add__(self, other):
        if isinstance(other, Number):
            q = self.duplicate()
            q[(0,) * self.dim] += other

            return q
        elif not isinstance(other, ComplexL0SequenceMD):
            raise TypeError("Sequence addition admits only other sequences or scalars.")
        
        return self._coeffwise_binary(other, lambda x, y: x + y)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return self._coeffwise_unary(lambda x: -x)
    
    def __sub__(self, other):
        if isinstance(other, Number):
            q = self.duplicate()
            q[(0,) * self.dim] -= other

            return q
        elif not isinstance(other, ComplexL0SequenceMD):
            raise TypeError("Sequence addition admits only other sequences or scalars.")
        
        return self._coeffwise_binary(other, lambda x, y: x - y)
    
    def __rsub__(self, other):
        if isinstance(other, Number):
            return other + (-self)
        elif not isinstance(other, ComplexL0SequenceMD):
            raise TypeError("Sequence addition admits only other sequences or scalars.")
        
        return other - self
    


class PolynomialMD(ComplexL0SequenceMD):
    """Represents a general multivariate Laurent polynomial."""

    def __init__(self, coeffs: list, support_start: tuple[int] | int):
        """Initializes a PolynomialMD instance.

        Args:
            coeffs: List of complex numbers as coefficients.
            support_start: tuple of minimum degrees in the polynomial.
        """
        super().__init__(coeffs, support_start)
    
    def shift(self, k: int, a: int=0):
        """Creates a new polynomial equal to the current one, multiplied by `z_a^k`."""
        t = self.support_start
        t = tuple(t[j] + (j == a) * k for j in range(len(t)))

        return PolynomialMD(self.coeff_list(), t)

    def effective_degree(self) -> int:
        """Returns a tuple containing the effective degrees with respect to each variable (max degree - min degree).

        Note:
            This does not check for leading or trailing zeros in the coefficient array.
        """
        t = self.support()
        return tuple(rng.stop - rng.start - 1 for rng in t)

    def conjugate(self):
        r"""Returns the conjugate polynomial on the unit circle. If :math:`p(z) = \sum_k p_k z^k`, then its conjugate is defined as :math:`p^*(z) = \sum_k p_k^* z^{-k}`

        Returns:
            PolynomialMD: The conjugate polynomial.
        """
        cf = self.coeff_list()
        deep_inplace(cf, lambda x: bd.conj(x), reverse=True)

        return PolynomialMD(cf, tuple(-rng.stop + 1 for rng in self.support()))
    
    def schwarz_transform(self):
        r"""Returns the anti-analytic polynomial whose real part gives the current polynomial.
        
        In other words, this is equivalent to adding :math:`iH[p]`, where :math:`H[p]` is the Hilbert transform of p.

        Returns:
            Polynomial: The Schwarz transform of the polynomial.
        """
        st = self.duplicate() # TODO This method can be more efficient

        for k in itertools.product(*self.support()):
            st[k] = schwarz_multiplier(k) * st[k]

        return st
    
    def __mul__(self, other):
        if isinstance(other, Number):
            return self._coeffwise_unary(lambda x: x * other)
        elif not isinstance(other, PolynomialMD):
            raise TypeError("Polynomial addition admits only other polynomials or scalars.")
        
        sup_a = self.support()
        sup_b = other.support()

        len_a = tuple(x.stop - x.start for x in sup_a)
        len_b = tuple(x.stop - x.start for x in sup_b)
        len_c = tuple(la + lb - 1 for la, lb in zip(len_a, len_b))

        rng_a = tuple(range(xa.start, xa.start + next_power_of_two(xc)) for xa, xc in zip(sup_a, len_c))
        rng_b = tuple(range(xb.start, xb.start + next_power_of_two(xc)) for xb, xc in zip(sup_b, len_c))
        # augmented support for a and b so that we can carry out FFT on their coeff_list()

        # TODO use extra precision here
        cf1 = bd.fft_md(self.coeff_list(rng_a))
        cf2 = bd.fft_md(other.coeff_list(rng_b))

        # Multiply in the Fourier domain
        deep_inplace_binary(cf1, cf2, lambda x, y: x * y) # cf1 *= cf2

        # Inverse FFT to get the result, support starts are the sum in each individual variable
        return PolynomialMD(deep_truncate(bd.ifft_md(cf1), len_c), tuple(x.start + y.start for x, y in zip(sup_a, sup_b)))
    
    def __rmul__(self, other):
        return self * other
    
    def __truediv__(self, other):
        if isinstance(other, Number):
            return self._coeffwise_unary(lambda x: x / other)
        
        raise TypeError("Polynomial division is only possible with scalars.")
    
    def __call__(self, *z) -> generic_complex:
        """Evaluates the polynomial using Horner's method.

        Args:
            z (complex): The point at which to evaluate the polynomial.

        Returns:
            complex: The evaluated result.
        """
        if len(z) == 1 and isinstance(z[0], tuple):
            z = z[0]

        if self.dim != len(z):
            raise ValueError("Incorrect number of variables to evaluate the multivariate polynomial.")

        return deep_horner_eval(self.coeff_list(), z) * prod(zk ** nk for zk, nk in zip(z, self.support_start))

    def eval_at_roots_of_unity(self, N: int | tuple) -> list[generic_complex]:
        """Evaluates the polynomial at the N-th roots of unity using the inverse FFT.

        Args:
            N (int | tuple): A power of two specifying the number of roots. If N is not a power of two, then the next power of two is taken.
                             A tuple of appropriate dimension can be given, where the k-th variable will be computed in the `N[k]`-th roots of unity.

        Returns:
            list[complex]: List of evaluations at the N-th roots of unity.
            The k-th element will be `self[w^k]`, where `w` is the main N-th root of unity.
        """
        if isinstance(N, Number):
            N = (N,) * self.dim

        if self.dim != len(N):
            raise ValueError("The dimension of the grid must match the dimension of the polynomial.")

        sup = self.support()

        N = tuple(next_power_of_two(k) for k in N)
        M = tuple(next_power_of_two(max(k, r.stop - r.start)) for k, r in zip(N, sup))

        coeffs = self.coeff_list(tuple(range(r.start, r.start + k) for k, r in zip(M, sup)))
        coeffs = deep_sequence_shift(coeffs, self.support_start)
        # This has the effect of having everything multiplied by z[k]^s[k] for each k

        evals = bd.ifft_md(coeffs, normalize=False) # M evaluations at the M-th roots of unity
        return deep_sequence_skip(evals, tuple(m//n for m, n in zip(M, N)))
    
    def sup_norm(self, N=1024):
        """Estimates the supremum norm of the polynomial over the unit multitorus.
        
        Args:
            N (int, optional): the number of samples to compute the maximum from. If N is not a power of two, then the next power of two is taken.

        Returns:
            float: An estimate for the supremum norm of the polynomial over the unit circle.
        """
        return max([abs(x) for x in flatten(self.eval_at_roots_of_unity(N))])
    
    def truncate(self, *rng: tuple[range] | tuple[tuple[range]]):
        """Keeps only the coefficients in the hyperrectangle defined by rng, discarding the other.
        
        Note:
            Both ends of the ranges are included.

        Args:
            rng (tuple[range]): Lower bound of degree.

        Returns:
            PolynomialMD: A new, truncated polynomial.
        """
        if len(rng) == 1 and isinstance(rng, tuple):
            rng = rng[0]
        
        rng = tuple(range(r.start, r.stop+1) for r in rng)

        return PolynomialMD(self.coeff_list(rng), tuple(r.start for r in rng))
    
    def __str__(self, tol=bd.machine_threshold()):
        """Converts the polynomial to a human-readable string representation.

        Returns:
            str: The string representation of the polynomial.
        """
        sup_start = self.support_start
        return ' + '.join(f"{c} z^{tuple(x + y for x, y in zip(k, sup_start))}" for k, c in self.coeff_dict().items() if bd.abs(c) > tol)
    

def to_poly_md(p: Polynomial, m: int) -> PolynomialMD:
    """Converts the given univariate polynomial P(z_m) as a m-variate polynomial object representing `P'(z_1, ..., z_m) = P(z_m)`."""

    cf = p.coeffs
    for _ in range(m-1):
        cf = [cf]

    return PolynomialMD(cf, support_start=(0,)*(m-1) + (p.support_start,))



import unittest

import nlft_qsp.numerics as bd
import numpy as np

from nlft_qsp.poly import Polynomial
from nlft_qsp.poly_matrix import MatrixPolynomial
from nlft_qsp.rand import random_sequence
from nlft_qsp.nlft import NonLinearFourierSequence


class MatrixPolynomialTestCase(unittest.TestCase):

    def test_init(self):
        """Test MatrixPolynomial initialization."""
        M = MatrixPolynomial((2, 3))
        self.assertEqual(M.shape, (2, 3))
        self.assertEqual(len(M._sequences), 0)

    def test_get_set_sequence(self):
        """Test getting and setting entire sequences."""
        M = MatrixPolynomial((2, 2))
        
        # Get returns empty sequence initially
        seq = M[0, 1]
        self.assertEqual(len(seq.coeffs), 0)
        
        # Set a sequence
        p = Polynomial([1, 2, 3], support_start=-1)
        M[0, 1] = p
        
        self.assertEqual(M[0, 1].coeffs, [1, 2, 3])
        self.assertEqual(M[0, 1].support_start, -1)

    def test_get_set_element(self):
        """Test getting and setting individual coefficients."""
        M = MatrixPolynomial((2, 2))
        
        # Set element at (0, 1, 2) -> coefficient of z^2 at position (0, 1)
        M[0, 1, 2] = 5 + 3j
        self.assertEqual(M[0, 1, 2], 5 + 3j)
        
        # Get element that doesn't exist
        self.assertEqual(M[1, 1, 5], bd.make_complex(0))
        
        # Set multiple elements in same sequence
        M[0, 1, -1] = 2 + 1j
        self.assertEqual(M[0, 1, -1], 2 + 1j)
        self.assertEqual(M[0, 1, 2], 5 + 3j)

    def test_get_matrix_at_degree(self):
        """Test getting the full matrix at a specific degree."""
        M = MatrixPolynomial((2, 2))
        
        M[0, 0, 0] = 1
        M[0, 1, 0] = 2
        M[1, 0, 0] = 3
        M[1, 1, 0] = 4
        
        mat = M[:, :, 0]
        self.assertEqual(mat, [[1, 2], [3, 4]])
        
        # Degree with no entries
        mat = M[:, :, 5]
        self.assertEqual(mat, [[0, 0], [0, 0]])

    def test_add_matrix_sequences(self):
        """Test addition of two matrix polynomials."""
        M1 = MatrixPolynomial((2, 2))
        M2 = MatrixPolynomial((2, 2))
        
        M1[0, 0, 0] = 1
        M1[1, 1, 1] = 2
        
        M2[0, 0, 0] = 3
        M2[0, 1, 1] = 4
        
        M3 = M1 + M2
        self.assertEqual(M3[0, 0, 0], 4)
        self.assertEqual(M3[0, 1, 1], 4)
        self.assertEqual(M3[1, 1, 1], 2)

    def test_add_constant_matrix(self):
        """Test adding a constant matrix to matrix polynomial."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 1] = 5
        
        const = [[1, 2], [3, 4]]
        M2 = M + const
        
        # Constant is added to 0-th coefficient
        self.assertEqual(M2[0, 0, 0], 1)
        self.assertEqual(M2[0, 1, 0], 2)
        self.assertEqual(M2[1, 0, 0], 3)
        self.assertEqual(M2[1, 1, 0], 4)
        
        # Original polynomial part is preserved
        self.assertEqual(M2[0, 0, 1], 5)

    def test_negation(self):
        """Test negation of matrix polynomial."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 2 + 1j
        M[1, 1, -1] = 3 - 2j
        
        M_neg = -M
        self.assertEqual(M_neg[0, 0, 0], -2 - 1j)
        self.assertEqual(M_neg[1, 1, -1], -3 + 2j)

    def test_subtraction(self):
        """Test subtraction of matrix polynomials."""
        M1 = MatrixPolynomial((2, 2))
        M2 = MatrixPolynomial((2, 2))
        
        M1[0, 0, 0] = 5
        M2[0, 0, 0] = 2
        
        M3 = M1 - M2
        self.assertEqual(M3[0, 0, 0], 3)

    def test_scalar_multiplication(self):
        """Test multiplication by scalar."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 2
        M[1, 1, 1] = 3
        
        M2 = M * 5
        self.assertEqual(M2[0, 0, 0], 10)
        self.assertEqual(M2[1, 1, 1], 15)
        
        M3 = 2 * M
        self.assertEqual(M3[0, 0, 0], 4)
        self.assertEqual(M3[1, 1, 1], 6)

    def test_scalar_division(self):
        """Test division by scalar."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 10
        M[1, 1, 1] = 6
        
        M2 = M / 2
        self.assertEqual(M2[0, 0, 0], 5)
        self.assertEqual(M2[1, 1, 1], 3)

    @bd.workdps(30)
    def test_matrix_multiplication(self):
        """Test matrix polynomial multiplication via pointwise evaluation.

        Build two random MatrixPolynomial instances of degree 5 and check
        M1(z) @ M2(z) == (M1 * M2)(z) for several random z.
        """
        import random

        deg = 5
        shape = (2, 2)

        M1 = MatrixPolynomial(shape)
        M2 = MatrixPolynomial(shape)

        # populate with random polynomials (support_start = 0)
        for i in range(shape[0]):
            for j in range(shape[1]):
                M1[i, j] = Polynomial(random_sequence(1.0, deg + 1), support_start=0)
                M2[i, j] = Polynomial(random_sequence(1.0, deg + 1), support_start=0)

        M3 = M1 * M2

        trials = 8
        for _ in range(trials):
            z = complex(random.random() * 2 - 1, random.random() * 2 - 1)
            A = M1(z)
            B = M2(z)
            C = M3(z)

            # compute A @ B using numpy
            A_np = np.array(A, dtype=complex)
            B_np = np.array(B, dtype=complex)
            prod_np = A_np @ B_np
            C_np = np.array(C, dtype=complex)

            # compare entrywise
            tol = 10 * bd.machine_threshold()
            for i in range(shape[0]):
                for j in range(shape[1]):
                    self.assertAlmostEqual(prod_np[i, j], C_np[i, j], delta=tol)

    @bd.workdps(30)
    def test_matrix_call(self):
        """Test evaluating matrix polynomial at a point."""
        M = MatrixPolynomial((2, 2))
        
        # M = [[1 + z, 0], [0, 2 - z]]
        M[0, 0, 0] = 1
        M[0, 0, 1] = 1
        M[1, 1, 0] = 2
        M[1, 1, 1] = -1
        
        result = M(2 + 1j)
        
        # Expected: [[1 + (2+1j), 0], [0, 2 - (2+1j)]] = [[3+1j, 0], [0, -1j]]
        self.assertAlmostEqual(result[0][0], 3 + 1j, delta=1e-10)
        self.assertAlmostEqual(result[0][1], 0, delta=1e-10)
        self.assertAlmostEqual(result[1][0], 0, delta=1e-10)
        self.assertAlmostEqual(result[1][1], -1j, delta=1e-10)

    @bd.workdps(30)
    def test_eval_at_roots_of_unity(self):
        """Test evaluation at roots of unity."""
        deg = 5
        shape = (2, 2)

        M = MatrixPolynomial(shape)
        for i in range(shape[0]):
            for j in range(shape[1]):
                M[i, j] = Polynomial(random_sequence(1.0, deg + 1), support_start=0)
                M[i, j] = Polynomial(random_sequence(1.0, deg + 1), support_start=0)

        N = 16
        ep = M.eval_at_roots_of_unity(N)
        cep = [M(z) for z in bd.unitroots(N)]

        # Check that each evaluation is a 2x2 matrix
        for mat in ep:
            self.assertEqual(len(mat), 2)
            self.assertEqual(len(mat[0]), 2)

        ep_np = [np.array(mat, dtype=complex) for mat in ep]
        cep_np = [np.array(mat, dtype=complex) for mat in cep]

        tol = 10 * bd.machine_threshold()
        for k in range(len(ep)):
            for i in range(M.shape[0]):
                for j in range(M.shape[1]):
                    self.assertAlmostEqual(ep_np[k][i][j], cep_np[k][i][j], delta=tol)

    def test_duplicate(self):
        """Test duplication of matrix polynomial."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 1 + 2j
        M[1, 1, -1] = 3
        
        M2 = M.duplicate()
        
        self.assertEqual(M2[0, 0, 0], 1 + 2j)
        self.assertEqual(M2[1, 1, -1], 3)
        
        # Modify original, duplicate should be unaffected
        M[0, 0, 0] = 5
        self.assertEqual(M2[0, 0, 0], 1 + 2j)

    def test_shift(self):
        """Test shifting matrix polynomial by power of z."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 1
        M[0, 0, 1] = 2
        
        M2 = M.shift(3)
        
        # Degrees should shift by 3
        self.assertEqual(M2[0, 0, 3], 1)
        self.assertEqual(M2[0, 0, 4], 2)
        self.assertEqual(M2[0, 0, 1], bd.make_complex(0))

    def test_conjugate(self):
        """Test conjugation of matrix polynomial."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, 0] = 1 + 2j
        M[0, 0, 1] = 3 - 1j
        
        M_conj = M.conjugate()
        
        self.assertEqual(M_conj[0, 0, 0], 1 - 2j)
        self.assertEqual(M_conj[0, 0, -1], 3 + 1j)

    def test_truncate(self):
        """Test truncation of matrix polynomial."""
        M = MatrixPolynomial((2, 2))
        M[0, 0, -2] = 1
        M[0, 0, 0] = 2
        M[0, 0, 2] = 3
        
        M2 = M.truncate(-1, 1)
        
        self.assertEqual(M2[0, 0, -2], bd.make_complex(0))
        self.assertEqual(M2[0, 0, 0], 2)
        self.assertEqual(M2[0, 0, 2], bd.make_complex(0))

    def test_shape_validation(self):
        """Test shape validation."""
        M1 = MatrixPolynomial((2, 3))
        M2 = MatrixPolynomial((3, 2))
        
        # Should work: 2x3 * 3x2
        M3 = M1 * M2
        self.assertEqual(M3.shape, (2, 2))
        
        # Should fail: 2x3 + 3x2
        with self.assertRaises(ValueError):
            M1 + M2


if __name__ == '__main__':
    unittest.main()



from .. import numerics as bd

from ..solvers.weiss import WEISS_MAX_ATTEMPTS, WeissConvergenceError
from ..util import next_power_of_two
from ..poly_md import PolynomialMD, deep_inplace, deep_sequence_shift

def laurent_approximation_md(points: list, m: int) -> PolynomialMD:
    r"""Returns a Laurent polynomial passing through the given points.

    Note:
        `N = len(points)` is assumed to be a power of two.

    Args:
        points (list[complex]): list of values, where the (k_1, ..., k_m)-th element is considered to be :math:`f(e^{2\pi i k_1/N}, ..., e^{2\pi i k_m/N})`.
        m (int): dimension of points, i.e., number of variables for the Laurent approximation.

    Returns:
        PolynomialMD: The unique Laurent polynomial `P(z_1, ..., z_m)` of degree `N = len(points)` satisfying :math:`P(e^{2\pi i k_1/N}, ..., e^{2\pi i k_m/N}) = f(e^{2\pi i k_1/N}, ..., e^{2\pi i k_m/N})`, up to working precision, whose frequencies are shifted to be in :math:`[-N/2, N/2)^m`
    """
    N = len(points)
    support_start = (-N//2,) * m

    coeffs = bd.fft_md(points, normalize=True)
    coeffs = deep_sequence_shift(coeffs, support_start) # Zero frequency in the middle
    
    return PolynomialMD(coeffs, support_start)

def complete_md(b: PolynomialMD, verbose=False):
    """Uses the Weiss-Schwarz algorithm to find a complementary polynomial to the given one. The polynomial will also be the unique outer, positive-mean polynomial with this property.

    Args:
        b (PolynomialMD): The polynomial to complete.
        verbose (bool, optional): verbosity during the procedure. Defaults to False.

    Returns:
        PolynomialMD: A polynomial :math:`a(z)` satisfying :math:`|a|^2 + |b|^2 = 1` on the unit circle (up to working precision).
    """
    N = 4*next_power_of_two(max(k for k in b.effective_degree())) # Exponential search on N, this can be better.
    m = b.dim # Number of variables

    threshold = 1
    attempts = 0
    while threshold > 100 * bd.machine_eps():
        N *= 2

        b_points = b.eval_at_roots_of_unity(N)
        deep_inplace(b_points, lambda bz: bd.log(1 - bd.abs2(bz))/2)
        R = laurent_approximation_md(b_points, m)

        G = R.schwarz_transform()

        G_points = G.eval_at_roots_of_unity(N)
        deep_inplace(G_points, lambda gz: bd.exp(gz))
        a = laurent_approximation_md(G_points, m)

        #a = a.truncate(-b.effective_degree(), 0) # a and b must have the same support

        new_thr = (a * a.conjugate() + b * b.conjugate() - 1).l2_norm()
        if verbose:
            print(f"N = {N:>7}, threshold = {new_thr}")

        if threshold <= new_thr:
            attempts += 1
            if attempts >= WEISS_MAX_ATTEMPTS:
                raise WeissConvergenceError()
        else:
            threshold = new_thr
            attempts = 0

    return a



from numbers import Number

from . import numerics as bd
from .numerics import coeffs_pad
from .numerics.backend import generic_complex, generic_real

from .util import next_power_of_two, sequence_shift


class ComplexL0Sequence:
    """Represents a sequence of complex numbers index by Z, whose support is finite.
    
    Attributes:
        coeffs (list[generic_complex]): List of complex coefficients.
        support_start (int): Index of the first element of the sequence.
    """

    def __init__(self, coeffs: list[generic_complex], support_start: int = 0):
        """Initializes a complex sequence.

        Args:
            coeffs: List of complex numbers as coefficients.
            support_start (optional): Index of the first element of the sequence. Defaults to 0.
        """
        self.coeffs = [bd.make_complex(c) for c in coeffs]
        self.support_start = support_start

    def support(self) -> range:
        """Returns the range in Z where the sequence is non-zero.

        Note:
            This simply checks the allocated array of coefficients but does not check leading or trailing zeros.

        Returns:
            range: The support of the sequence.
        """
        return range(self.support_start, self.support_start + len(self.coeffs))
    
    def __getitem__(self, k: int) -> generic_complex:
        """Returns the k-th element of the sequence, i.e., F_k.

        Args:
            k (int): The index of the sequence.

        Returns:
            complex: The coefficient of F_k, or 0 if k is out of the support.
        """
        if k in self.support():
            return self.coeffs[k - self.support_start]
        return bd.make_complex(0)

    def __setitem__(self, k: int, c: generic_complex):
        """Sets the coefficient of z^k to be c, allocating space if needed.

        Args:
            k (int): The exponent of z.
            c (complex): The coefficient to set.
        """
        if self.support_start + len(self.coeffs) <= k:
            self.coeffs.extend([bd.make_complex(0)] * (k - self.support_start - len(self.coeffs) + 1))
        elif self.support_start > k:
            self.coeffs = [bd.make_complex(0)] * (self.support_start - k) + self.coeffs
            self.support_start = k
        self.coeffs[k - self.support_start] = bd.make_complex(c)

    def l1_norm(self) -> generic_real:
        """Computes the l1 norm of the sequence.

        Returns:
            float: The sum of absolute values of coefficients.
        """
        return sum(bd.abs(c) for c in self.coeffs)

    def l2_norm(self) -> generic_real:
        """Computes the l2 norm.

        Returns:
            float: The l2 norm.
        """
        return bd.sqrt(self.l2_squared_norm())

    def l2_squared_norm(self) -> generic_real:
        """Computes the squared l2 norm.

        Returns:
            float: The squared l2 norm, i.e., the sum of the squared absolute values.
        """
        return sum(bd.abs2(c) for c in self.coeffs)
    
    def is_real(self) -> bool:
        """Whether the sequence has only real elements."""
        return all(bd.im(F) <= bd.machine_threshold() for F in self.coeffs)
    
    def is_imaginary(self) -> bool:
        """Whether the sequence has only imaginary elements."""
        return all(bd.re(F) <= bd.machine_threshold() for F in self.coeffs)
    
    def is_symmetric(self) -> bool:
        """Whether the sequence satisfies F[k] = F[-k]."""
        for k in self.support():
            if abs(self[k] - self[-k]) > bd.machine_threshold():
                return False
        return True
    
    def __add__(self, other):
        if isinstance(other, Number):
            q = self.duplicate()
            q[0] += other

            return q
        elif not isinstance(other, Polynomial):
            raise TypeError("Polynomial addition admits only other polynomials or scalars.")
                
        self_end = self.support_start + len(self.coeffs)
        other_end = other.support_start + len(other.coeffs)
        
        sum_start = min(self.support_start, other.support_start)
        sum_end = max(self_end, other_end)

        sum_coeffs = []
        for k in range(sum_start, sum_end):
            res = bd.make_complex(0)
            
            if self.support_start <= k and k < self_end:
                res += self.coeffs[k - self.support_start]

            if other.support_start <= k and k < other_end:
                res += other.coeffs[k - other.support_start]

            sum_coeffs.append(res)
            
        return Polynomial(sum_coeffs, sum_start)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return Polynomial([-c for c in self.coeffs], self.support_start)
    
    def __sub__(self, other):
        return self + (-other)
    
    def __rsub__(self, other):
        return self + (-other)


class Polynomial(ComplexL0Sequence):
    """Represents a general Laurent polynomial of one complex variable.

    Attributes:
        coeffs (list[generic_complex]): List of complex coefficients.
        support_start (int): Minimum degree that appears in the polynomial.
    """

    def __init__(self, coeffs: list[generic_complex], support_start: int = 0):
        """Initializes a Polynomial instance.

        Args:
            coeffs: List of complex numbers as coefficients.
            support_start (optional): Minimum degree in the polynomial. Defaults to 0.
        """
        super().__init__(coeffs, support_start)

    def duplicate(self):
        """Creates a duplicate of the current polynomial.

        Returns:
            Polynomial: A new Polynomial instance with the same coefficients and support.
        """
        return Polynomial(self.coeffs, self.support_start)
    
    def shift(self, k: int):
        """Creates a new polynomial equal to the current one, multiplied by `z^k`."""
        return Polynomial(self.coeffs, self.support_start + k)

    def effective_degree(self) -> int:
        """Returns the size of the support of the polynomial minus 1 (max degree - min degree).

        Note:
            This does not check for leading or trailing zeros in the coefficient array.

        Returns:
            int: The effective degree of the polynomial.
        """
        return len(self.coeffs) - 1

    def conjugate(self):
        r"""Returns the conjugate polynomial on the unit circle. If :math:`p(z) = \sum_k p_k z^k`, then its conjugate is defined as :math:`p^*(z) = \sum_k p_k^* z^{-k}`

        Returns:
            Polynomial: The conjugate polynomial.
        """
        conj_coeffs = [bd.conj(x) for x in reversed(self.coeffs)]
        return Polynomial(conj_coeffs, -(self.support_start + len(self.coeffs) - 1))
    
    def sharp(self):
        r"""Same as `conjugate()`, but the support_start is left unchanged.

        Returns:
            Polynomial: The sharp-conjugate polynomial.
        """
        p = self.conjugate()
        p.support_start += self.effective_degree() + 1
        return p

    def schwarz_transform(self):
        r"""Returns the anti-analytic polynomial whose real part gives the current polynomial.
        
        In other words, this is equivalent to adding :math:`iH[p]`, where :math:`H[p]` is the Hilbert transform of p.

        Returns:
            Polynomial: The Schwarz transform of the polynomial.
        """
        schwarz_coeffs = []
        for k in self.support():
            if k < 0:
                schwarz_coeffs.append(2*self[k])
            elif k == 0:
                schwarz_coeffs.append(self[k])

        return Polynomial(schwarz_coeffs, self.support_start)

    def __mul__(self, other):
        if isinstance(other, Number):
            return Polynomial([other * c for c in self.coeffs], self.support_start)
        elif not isinstance(other, Polynomial):
            raise TypeError("Polynomial addition admits only other polynomials or scalars.")
        len_c = len(self.coeffs) + len(other.coeffs) - 1

        # TODO use extra precision here
        coeffs_a = bd.fft(coeffs_pad(self.coeffs, next_power_of_two(len_c)))
        coeffs_b = bd.fft(coeffs_pad(other.coeffs, next_power_of_two(len_c)))

        # Multiply in the Fourier domain
        coeffs_c = [a * b for a, b in zip(coeffs_a, coeffs_b)]

        # Inverse FFT to get the result
        new_coeffs = bd.ifft(coeffs_c)
        support_start = self.support_start + other.support_start  # Lowest degree of the new poly

        return Polynomial(new_coeffs[0:len_c], support_start)
    
    def __rmul__(self, other):
        return self * other
    
    def __truediv__(self, other):
        if isinstance(other, Number):
            return Polynomial([c / other for c in self.coeffs], self.support_start)
        
        raise TypeError("Polynomial division is only possible with scalars.")

    def __call__(self, z) -> generic_complex:
        """Evaluates the polynomial using Horner's method.

        Args:
            z (complex): The point at which to evaluate the polynomial.

        Returns:
            complex: The evaluated result.
        """
        res = self.coeffs[-1]
        for k in reversed(range(len(self.coeffs) - 1)):
            res = res * z + self.coeffs[k]
        return res * (z ** self.support_start)

    def eval_at_roots_of_unity(self, N: int) -> list[generic_complex]:
        """Evaluates the polynomial at the N-th roots of unity using the inverse FFT.

        Args:
            N (int): A power of two specifying the number of roots. If N is not a power of two, then the next power of two is taken.

        Returns:
            list[complex]: List of evaluations at the N-th roots of unity.
            The k-th element will be `self[w^k]`, where `w` is the main N-th root of unity.
        """
        N = next_power_of_two(N)
        M = next_power_of_two(max(N, len(self.coeffs)))

        coeffs = coeffs_pad(self.coeffs, M)
        coeffs = sequence_shift(coeffs, self.support_start)
        # This has the effect of having everything multiplied by z^s

        evals = bd.ifft(coeffs, normalize=False) # M evaluations at the M-th roots of unity
        return evals[::M//N]
    
    def sup_norm(self, N=1024):
        """Estimates the supremum norm of the polynomial over the unit circle
        
        Args:
            N (int, optional): the number of samples to compute the maximum from. If N is not a power of two, then the next power of two is taken.

        Returns:
            float: An estimate for the supremum norm of the polynomial over the unit circle.
        """
        return max([abs(sample) for sample in self.eval_at_roots_of_unity(N)])
    
    def truncate(self, m: int, n: int):
        """Keeps only the coefficients in [m, n], discarding the others.

        Args:
            m (int): Lower bound of degree.
            n (int): Upper bound of degree.

        Returns:
            Polynomial: A new, truncated polynomial.
        """
        return Polynomial([self[k] for k in range(m, n+1)], m)
    
    def only_positive_degrees(self):
        """Discards all the negative degrees, keeping only the non-negative ones.
        
        Returns:
            Polynomial: A new polynomial containing only the positive-degree coefficients."""
        return self.truncate(0, self.support_start + len(self.coeffs) - 1)

    def __str__(self):
        """Converts the polynomial to a human-readable string representation.

        Returns:
            str: The string representation of the polynomial.
        """
        return ' + '.join(f"{c} z^{self.support_start + k}" for k, c in enumerate(self.coeffs))
    

class ChebyshevTExpansion(ComplexL0Sequence):
    """Linear combination of Chebyshev polynomials of the first kind.
    
    Args:
        c: Either the coefficients of the linear combination, or the symmetric Laurent polynomial `P(z)` which are equal up to the change of variable `x = (z + z^(-1))/2`.
    """
    def __init__(self, c: list[generic_complex] | Polynomial):
        if isinstance(c, list):
            super().__init__(c, support_start=0)
        elif isinstance(c, Polynomial):
            if not c.is_symmetric():
                raise ValueError("The given Laurent polynomial is not symmetric.")

            coeffs = [2*c[k] for k in range(c.support().stop)]
            coeffs[0] /= 2
            super().__init__(coeffs, support_start=0)
        else:
            raise ValueError("Only a coefficient vector or symmetric Laurent polynomials are allowed.")
        
    def degree(self) -> int:
        return len(self.coeffs) - 1

    def __call__(self, x: generic_real) -> generic_complex:
        """Evaluates the Chebyshev expansion at the given number.

        Args:
            x (real): The point at which to evaluate the expansion.

        Returns:
            complex: The evaluated result.
        """
        theta = bd.arccos(x)

        return sum(self[k] * bd.cos(k * theta) for k in self.support())
    
    def to_laurent(self):
        """Returns the Laurent polynomial `P(z) = self((z + z^(-1))/2)`."""
        P = Polynomial(list(reversed(self.coeffs)) + self.coeffs[1:], support_start=-len(self.coeffs)+1)
        P[0] *= 2
        return P/2
    
    @classmethod
    def from_polynomial(cls, P: Polynomial):
        """Returns the Chebyshev expansion `T` satisfying `T(x) = P(x)`."""
        return ChebyshevTExpansion(bd.poly2cheb(P.coeffs))
    
    @classmethod
    def from_laurent_polynomial(cls, P: Polynomial):
        """Returns the Chebyshev expansion `T` satisfying `T(x) = (P(z) + P^*(z))/2`.
        
        Note: `P` must be symmetric."""
        if not P.is_symmetric():
            raise ValueError("The given Laurent polynomial is not symmetric.")

        coeffs = [2*P[k] for k in range(P.support().stop)]
        coeffs[0] /= 2

        return ChebyshevTExpansion(coeffs)
    
    def to_polynomial(self) -> Polynomial:
        """Returns the polynomial `P` satisfying `P(x) = T(x)`."""
        return Polynomial(bd.cheb2poly(self.coeffs))


# Plot functions

from matplotlib import pyplot as plt

from . import numerics as bd

def plot_chebyshev(funcs: dict, num_points: int=1000):
    """
    Plots the real part of each object in funcs over the interval [-1, 1].
    These can be Python functions, `Polynomial` objects, `ChebyshevTExpansion` objects, or
    any callable object.

    Parameters:
    - funcs (dict): a dictionary where each key is the name appearing in the legend of the corresponding function plot.
    - num_points (int): number of sampling points.
    """
    plt.figure(figsize=(6, 3))
    
    for name, f in funcs.items():
        try:
            x_vals = [-1 + 2*k/num_points for k in range(num_points+1)]
            y_vals = [bd.re(f(x)) for x in x_vals]
            plt.plot(x_vals, y_vals, label=name)
        except Exception as e:
            print(f"Error evaluating function {name}: {e}")
    
    plt.xlabel("x")
    plt.legend()
    plt.grid(True)
    plt.show()

def plot_fourier(funcs: dict, num_points: int=1000):
    """
    Plots the absolute value of each object in funcs over the unit circle, i.e., plugging
    :math:`z = exp(1j*x)` for :math:`x \in [-\pi, \pi]`.
    These can be Python functions, `Polynomial` objects, or any callable object.

    Parameters:
    - funcs (dict): a dictionary where each key is the name appearing in the legend of the corresponding function plot.
    - num_points (int): number of sampling points.
    """
    plt.figure(figsize=(6, 3))
    
    for name, f in funcs.items():
        try:
            x_vals = [-bd.pi() + 2*bd.pi()*k/num_points for k in range(num_points+1)]
            y_vals = [bd.abs(f(bd.exp(1j * x))) for x in x_vals]
            plt.plot(x_vals, y_vals, label=name)
        except Exception as e:
            print(f"Error evaluating function {name}: {e}")
    
    plt.xlabel("x")
    plt.xlim(left=-bd.pi(), right=bd.pi())
    plt.legend()
    plt.grid(True)
    plt.show()

def plot_support_2d(l: list, rng: tuple[range]):
    """Plots where the given objects have non-zero elements (up to machine threshold) in rng.
    
    Args:
        l (list): A list of objects that support subscript operation.
        rng (tuple[range]): A 2D tuple of ranges defining the rectangle in the Z^2 grid to plot."""

    if len(rng) != 2:
        raise ValueError("rng must have dimension 2.")

    for k in range(len(l)):
        px, py = zip(*[
            (x+k*0.05, y+k*0.05) for x in rng[0] for y in rng[1] if bd.abs(l[k][x, y]) > bd.machine_threshold()
        ])
        
        plt.scatter(px, py, marker='o', label=f"Support #{k+1}")

    # Formatting
    plt.xlabel("k")
    plt.ylabel("h")
    plt.title("Supports")
    plt.grid(False)
    plt.legend()
    plt.gca().set_aspect('equal')

    plt.show()


import numpy as np

from .poly import ChebyshevTExpansion, Polynomial
from .util import next_power_of_two


def chebyshev_approximate(f, N) -> ChebyshevTExpansion:
    """
    Computes the Chebyshev expansion up to `N` for a complex-valued function f on [-1, 1].

    Args:
        f (callable): complex-valued function f(x)
        N (int): degree of Chebyshev approximation
    """
    x = np.cos(np.pi * np.arange(N + 1) / N)
    fx = [f(xk) for xk in x]

    # Correct mirrored extension
    fx_m = fx + fx[1:-1][::-1]

    # FFT without normalization
    F = np.fft.fft(fx_m)
    F = F[:N + 1] / N
    F[0]  /= 2
    F[-1] /= 2

    return ChebyshevTExpansion(F.tolist())

def fourier_approximate(f, N) -> Polynomial:
    r"""Computes the Fourier series of the given function `f(z)`

    Args:
        f (callable): a function taking a complex number z and returning a complex number.
        N (int): The degree of approximation.

    Returns:
        Polynomial: A Laurent polynomial of degrees in `[-N, N)` approximating f
    """
    M = next_power_of_two(2*N+1)

    zk = np.exp(2j * np.pi * np.arange(M) / M)
    fk = np.array([f(z) for z in zk], dtype=complex)

    pk = (np.fft.fft(fk) / M).tolist()
    
    return Polynomial(pk[M-N:] + pk[:N+1], support_start=-N)



from math import sqrt
import unittest

import nlft_qsp.numerics as bd

from nlft_qsp.nlft_md import StairlikeSequence2D
from nlft_qsp.poly_md import PolynomialMD
from nlft_qsp.rand import random_complex, random_list, random_sequence, random_stairlike_sequence_2d
from nlft_qsp.poly import Polynomial
from nlft_qsp.solvers import weiss_md


class PolynomialMDTestCase(unittest.TestCase):

    def test_get_set(self):
        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6],
             [7, 8, 9]], support_start=(0, 0))
        
        self.assertEqual(p[0,0], 1)
        self.assertEqual(p[0,1], 2)
        self.assertEqual(p[0,2], 3)
        self.assertEqual(p[1,0], 4)
        self.assertEqual(p[1,1], 5)
        self.assertEqual(p[1,2], 6)
        self.assertEqual(p[2,0], 7)
        self.assertEqual(p[2,1], 8)
        self.assertEqual(p[2,2], 9)

        self.assertEqual(p[-2,7], 0)
        p[-2,7] = 15
        self.assertEqual(p[-2,7], 15)
        self.assertEqual(p[-3,5], 0)
        self.assertEqual(p[-1,7], 0)

        self.assertEqual(p[1,8], 0)
        p[1,8] = 72
        self.assertEqual(p[1,8], 72)
        
        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6],
             [7, 8, 9]], support_start=(3, 2))
        self.assertEqual(p[3, 2], 1)

    def test_support(self):
        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6]], support_start=(0, 0))
        
        r1, r2 = p.support()
        self.assertEqual(r1, range(0, 2))
        self.assertEqual(r2, range(0, 3))

        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6]], support_start=(0, 2))
        
        r1, r2 = p.support()
        self.assertEqual(r1, range(0, 2))
        self.assertEqual(r2, range(2, 5))

        p = PolynomialMD(
            [[[1, 2], [3, 4], [5, 6]],
             [[1, 2], [3, 4], [5, 6]]], support_start=(1, 4, 5)
        )
        r1, r2, r3 = p.support()
        self.assertEqual(r1, range(1, 3))
        self.assertEqual(r2, range(4, 7))
        self.assertEqual(r3, range(5, 7))

    def test_coeff_list(self):
        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6]], support_start=(0, 0))
        self.assertEqual(p.coeff_list(), [[1, 2, 3], [4, 5, 6]])

        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6]], support_start=(2, 3))
        self.assertEqual(p.coeff_list(), [[1, 2, 3], [4, 5, 6]])

    def test_support_start(self):
        p = PolynomialMD(
            [[[1, 2], [3, 4], [5, 6]],
             [[1, 2], [3, 4], [5, 6]]], support_start=(0, 0, 0)
        )
        self.assertEqual(p.support_start, (0, 0, 0))

        p[-2, 0, 0] = 3
        self.assertEqual(p.support_start, (-2, 0, 0))

        p[-5, -3, -2] = 5
        self.assertEqual(p.support_start, (-5, -3, -2))

    def test_effective_degree(self):
        p = PolynomialMD(
            [[[1, 2], [3, 4], [5, 6]],
             [[1, 2], [3, 4], [5, 6]]], support_start=(0, 0, 0)
        )

        self.assertEqual(p.effective_degree(), (1, 2, 1))

        p[0, 0, 5] = 7
        self.assertEqual(p.effective_degree(), (1, 2, 5))

    def test_conjugate(self):
        p = PolynomialMD([[[0]]], (0, 0, 0))
        p[1, 2, 3] = 72 + 7j
        p[-5, -3, 1] = 36 - 5j

        self.assertEqual(p[1, 2, 3], 72 + 7j)
        self.assertEqual(p[-5, -3, 1], 36 - 5j)
        q = p.conjugate()
        self.assertEqual(q[-1, -2, -3], 72 - 7j)
        self.assertEqual(q[5, 3, -1], 36 + 5j)

    def test_add(self):
        p = PolynomialMD(
            [[1, 2, 3],
             [4, 5, 6]], support_start=(0, 0))
        
        q = PolynomialMD(
            [[7, 8, 9],
             [1, 3, 6]], support_start=(1, 1))
        
        self.assertEqual((p+q).coeff_list(),
                         [[1,   2,   3, 0],
                          [4, 5+7, 6+8, 9],
                          [0, 1,     3, 6]])
        self.assertAlmostEqual(((p+q) - (q+p)).l2_norm(), 0, delta=bd.machine_threshold())
        
        self.assertEqual((p-q).coeff_list(),
                         [[1,   2,   3,  0],
                          [4, 5-7, 6-8, -9],
                          [0,  -1,  -3, -6]])
        self.assertAlmostEqual(((p-q) + (q-p)).l2_norm(), 0, delta=bd.machine_threshold())
        
        self.assertEqual((q+35).coeff_list(),
                         [[35, 0, 0, 0],
                          [0,  7, 8, 9],
                          [0,  1, 3, 6]])
        
    def test_mul(self):
        p = PolynomialMD([[0, 1], [1, 0]], support_start=(0,0)) # x + y
        q = PolynomialMD([[0, 1], [-1, 0]], support_start=(1,0)) # x^2 - xy

        r = p*q
        self.assertEqual(r.support_start, (1,0))
        self.assertEqual(r.coeff_list(), [
            [0, 0, 1],
            [0, 0, 0],
            [-1, 0, 0]
        ])

        p = PolynomialMD([[0, 1], [1, 7]], support_start=(0,3)) # y^3 (x + y + 7xy)
        q = PolynomialMD([[0, 1], [-1, 0]], support_start=(1,0)) # x^2 - xy
        # x y^3 (x + y + 7xy) (x - y) = x y^3 (x^2 - y^2 + 7x^2 y - 7xy^2)
        
        r = p*q
        self.assertEqual(r.support_start, (1,3))
        self.assertEqual(r.coeff_list(), [
            [0, 0, 1],
            [0, 0, 7],
            [-1, -7, 0]
        ])

    def test_eval(self):
        p = PolynomialMD([ # P(x, y) = 2x^2 + 3xy + y^2
            [0, 0, 1],
            [0, 3, 0],
            [2, 0, 0]
        ], (0,0))
        self.assertEqual(p(1, 2), 12)

        p = PolynomialMD([ # P(x, y) = y^2(2x^2 + 3xy + y^2)
            [0, 0, 1],
            [0, 3, 0],
            [2, 0, 0]
        ], (0,2))
        self.assertEqual(p(1, 2), 48)

        p = PolynomialMD([ # P(x, y, z) = x + yz - z^2
            [[0, 0, -1],
             [0, 1, 0]],
            [[1, 0, 0],
             [0, 0, 0]]
        ], (0,0,0))
        self.assertEqual(p(2, 3, 1), 4)

    def test_eval_at_roots_of_unity(self):
        seq = random_sequence(10000, (4, 4, 4))

        p = PolynomialMD(seq, (0, 0, 0))
        q = PolynomialMD(seq, (0, 2, 1)) # q = y^2 z p
        
        ep = p.eval_at_roots_of_unity((6, 7, 9))
        eq = q.eval_at_roots_of_unity((6, 7, 9))

        for i, x in zip(range(8), bd.unitroots(8)):
            for j, y in zip(range(8), bd.unitroots(8)):
                for k, z in zip(range(16), bd.unitroots(16)):
                    self.assertAlmostEqual(ep[i][j][k], p(x, y, z), delta=10 * bd.machine_threshold())
                    self.assertAlmostEqual(eq[i][j][k], q(x, y, z), delta=10 * bd.machine_threshold())
                    self.assertAlmostEqual(eq[i][j][k], (y ** 2) * z * ep[i][j][k], delta=10 * bd.machine_threshold())

class Polynomial1DToMDTestCase(unittest.TestCase):

    def test_eval(self):
        seq = random_sequence(10, 16)
        w = random_complex(10)
        y = random_complex(10)

        p1 = PolynomialMD([seq], support_start=(0,0))
        p2 = PolynomialMD([[c] for c in seq], support_start=(0,0))
        p3 = PolynomialMD([[seq]], support_start=(0,0,0))
        q = Polynomial(seq, support_start=0)

        for z in bd.unitroots(16):
            self.assertAlmostEqual(p1(w, z), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p2(z, w), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p3(y, w, z), q(z), delta=10*bd.machine_threshold())

        p1 = PolynomialMD([seq], support_start=(0,3))
        p2 = PolynomialMD([[c] for c in seq], support_start=(3,0))
        p3 = PolynomialMD([[seq]], support_start=(0,0,3))
        q = Polynomial(seq, support_start=3)

        for z in bd.unitroots(16):
            self.assertAlmostEqual(p1(w, z), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p2(z, w), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p3(y, w, z), q(z), delta=10*bd.machine_threshold())

        e1 = p1.eval_at_roots_of_unity(16)
        e2 = p2.eval_at_roots_of_unity(16)
        e3 = p3.eval_at_roots_of_unity(16)

        eq = q.eval_at_roots_of_unity(16)

        for k in range(16):

            for h in range(16):
                self.assertAlmostEqual(e1[h][k], eq[k], delta=10*bd.machine_threshold())
                self.assertAlmostEqual(e2[k][h], eq[k], delta=10*bd.machine_threshold())

                for j in range(16):
                    self.assertAlmostEqual(e3[j][h][k], eq[k], delta=10*bd.machine_threshold())

    def test_eval_at_roots_of_unity(self):
        seq = random_sequence(10, 16)

        e1 = PolynomialMD([seq], support_start=(0,3))               .eval_at_roots_of_unity(16)
        e2 = PolynomialMD([[c] for c in seq], support_start=(3,0))  .eval_at_roots_of_unity(16)
        e3 = PolynomialMD([[seq]], support_start=(0,0,3))           .eval_at_roots_of_unity(16)

        eq = Polynomial(seq, support_start=3).eval_at_roots_of_unity(16)

        for k in range(16):
            for h in range(16):
                self.assertAlmostEqual(e1[h][k], eq[k], delta=10*bd.machine_threshold())
                self.assertAlmostEqual(e2[k][h], eq[k], delta=10*bd.machine_threshold())

                for j in range(16):
                    self.assertAlmostEqual(e3[j][h][k], eq[k], delta=10*bd.machine_threshold())

    def test_add(self):
        seq1 = random_sequence(10, 16)
        seq2 = random_sequence(10, 16)

        p1_md = PolynomialMD([seq1], support_start=(0,0))
        q1_md = PolynomialMD([seq2], support_start=(0,-5))

        p2_md = PolynomialMD([[c] for c in seq1], support_start=(0,0))
        q2_md = PolynomialMD([[c] for c in seq2], support_start=(-5,0))

        p = Polynomial(seq1, support_start=0)
        q = Polynomial(seq2, support_start=-5)

        r1_md = p1_md + q1_md
        r2_md = p2_md + q2_md
        r = p + q

        w = random_complex(10)
        for z in bd.unitroots(1024):
            self.assertAlmostEqual(r1_md(w, z), r(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(r2_md(z, w), r(z), delta=10*bd.machine_threshold())

    def test_mul(self):
        seq1 = random_sequence(10, 16)
        seq2 = random_sequence(10, 16)

        p1_md = PolynomialMD([seq1], support_start=(0,0))
        q1_md = PolynomialMD([seq2], support_start=(0,-5))

        p2_md = PolynomialMD([[c] for c in seq1], support_start=(0,0))
        q2_md = PolynomialMD([[c] for c in seq2], support_start=(-5,0))

        p = Polynomial(seq1, support_start=0)
        q = Polynomial(seq2, support_start=-5)

        r1_md = p1_md * q1_md
        r2_md = p2_md * q2_md
        r = p * q

        w = random_complex(10)
        for z in bd.unitroots(1024):
            self.assertAlmostEqual(r1_md(w, z), r(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(r2_md(z, w), r(z), delta=10*bd.machine_threshold())

    def test_schwarz_transform(self):
        seq = random_sequence(10, 10)
        w = random_complex(10)
        y = random_complex(10)

        p1 = PolynomialMD([seq], support_start=(0,0))               .schwarz_transform()
        p2 = PolynomialMD([[c] for c in seq], support_start=(0,0))  .schwarz_transform()
        p3 = PolynomialMD([[seq]], support_start=(0,0,0))           .schwarz_transform()
        q = Polynomial(seq, support_start=0)                        .schwarz_transform()

        for z in bd.unitroots(16):
            self.assertAlmostEqual(p1(w, z), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p2(z, w), q(z), delta=10*bd.machine_threshold())
            self.assertAlmostEqual(p3(y, w, z), q(z), delta=10*bd.machine_threshold())



class StairlikeSequence2DTestCase(unittest.TestCase):

    def test_init_get(self):
        s = StairlikeSequence2D([[1, 2], [3], [4, 5, 6], [7, 8]])

        self.assertEqual(s.support_x(), range(0, 4))
        self.assertEqual(s.support_y(), range(0, 5))

        self.assertEqual(s[0,0], 1)
        self.assertEqual(s[0,1], 2)
        self.assertEqual(s[1,1], 3)
        self.assertEqual(s[2,1], 4)
        self.assertEqual(s[2,2], 5)
        self.assertEqual(s[2,3], 6)
        self.assertEqual(s[3,3], 7)
        self.assertEqual(s[3,4], 8)

        self.assertEqual(s[1,0], 0)
        self.assertEqual(s[-1,0], 0)

    def test_nlft_transform(self):
        nlft = StairlikeSequence2D([[1, 2], [3], [4]])
        a, b = nlft.transform()

        self.assertAlmostEqual(a[0,0],     1/(10*sqrt(17)),  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(a[0,-1],    -1/(5*sqrt(17)),  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(a[-1,0],    -9/(5*sqrt(17)),  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(a[-1,-1],   21/(10*sqrt(17)), delta=10*bd.machine_threshold())
        self.assertAlmostEqual(a[-2,0],    -4/(5*sqrt(17)),  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(a[-2,-1],   -2/(5*sqrt(17)),  delta=10*bd.machine_threshold())

        self.assertAlmostEqual(b[0,0],     1/(10*sqrt(17)),   delta=10*bd.machine_threshold())
        self.assertAlmostEqual(b[0,1],     1/(5*sqrt(17)),    delta=10*bd.machine_threshold())
        self.assertAlmostEqual(b[1,0],     -9/(5*sqrt(17)),   delta=10*bd.machine_threshold())
        self.assertAlmostEqual(b[1,1],     -21/(10*sqrt(17)), delta=10*bd.machine_threshold())
        self.assertAlmostEqual(b[2,0],     -4/(5*sqrt(17)),   delta=10*bd.machine_threshold())
        self.assertAlmostEqual(b[2,1],     2/(5*sqrt(17)),    delta=10*bd.machine_threshold())

        nlft = StairlikeSequence2D(random_stairlike_sequence_2d(10, shape=(4, 4)), support_start=(2, 1))
        a, b = nlft.transform()
        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=10*bd.machine_threshold())

    def test_random_multiplication(self):
        nlft1 = StairlikeSequence2D(random_stairlike_sequence_2d(10, shape=(4, 4)))
        nlft2 = StairlikeSequence2D(random_stairlike_sequence_2d(10, shape=(4, 4)), support_start=(15, 10))

        a1, b1 = nlft1.transform()
        a2, b2 = nlft2.transform()

        r = b1 * b2

        z = (random_complex(1), random_complex(1))
        self.assertAlmostEqual(b1(z) * b2(z), r(z), delta=10 * bd.machine_threshold())

    def test_schwarz_transform(self):
        p = PolynomialMD(random_list(1, (3, 3)), support_start=(-1, -1))
        q = p.schwarz_transform()

        self.assertAlmostEqual(p[-1,-1]*2, q[-1,-1], delta=10*bd.machine_threshold())
        self.assertAlmostEqual(p[-1,0]*2,  q[-1,0],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(0,          q[-1,1],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(p[0,-1]*2,  q[0,-1],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(p[0, 0],    q[0, 0],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(0,          q[1,-1],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(0,          q[0, 1],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(0,          q[1, 0],  delta=10*bd.machine_threshold())
        self.assertAlmostEqual(0,          q[1, 1],  delta=10*bd.machine_threshold())

    def test_laurent_approx_md(self):
        N = 16
        points = random_list(1, (N, N)) # points[k][h] = f(\omega^k, \omega^h)

        P = weiss_md.laurent_approximation_md(points, 2)

        self.assertAlmostEqual(max(abs(P(bd.exp(2j*bd.pi()*k/N), bd.exp(2j*bd.pi()*h/N)) - points[k][h])
                                   for k in range(N) for h in range(N)), 0,
                                   delta=10*bd.machine_threshold())


if __name__ == '__main__':
    unittest.main()




from .backend import NumericBackend
from .backend import generic_complex, generic_real

from ..util import next_power_of_two


class MPMathBackend(NumericBackend):
    """Numeric backend for the mpmath package. This exposes the mpmath implementation.
    
    Note:
        The mpmath package allows to compute with arbitrary precision, but its performance is limited.
    """

    def __init__(self, ctx):
        """Initializes an mpmath backend interface with the given mpmath context.
        
        Args:
            ctx (mpmath.MPContext): the context object exposed by mpmath. It can be `mpmath.mp`, `mpmath.fp`, or `mpmath.iv`.
        """
        self.ctx = ctx

    def __getattr__(self, item): # redirect any other function to the mpmath context
        return getattr(self.ctx, item)

    def pi(self):
        return self.ctx.pi

    def machine_eps(self):
        return 10 ** (-self.ctx.dps)
    
    def machine_threshold(self):
        return 10 ** (-self.ctx.dps//2)
    
    def workdps(self, x: int):
        return self.ctx.workdps(x)
    
    def workprec(self, x: int):
        return self.ctx.workprec(x)
    
    def extradps(self, x: int):
        return self.ctx.extradps(x)
    
    def extraprec(self, x: int):
        return self.ctx.extraprec(x)
    
    def chop(self, x: generic_complex):
        return self.ctx.chop(x)

    def make_complex(self, x: generic_complex):
        return self.ctx.mpc(x)
    
    def make_float(self, x: generic_real):
        return self.ctx.mpf(x)

    def abs(self, x: generic_complex):
        return self.ctx.fabs(x)
    
    def abs2(self, x: generic_complex):
        return self.ctx.re(x)**2 + self.ctx.im(x)**2
    
    def arcsin(self, x: generic_complex):
        return self.ctx.asin(x)
    
    def arccos(self, x: generic_complex):
        return self.ctx.acos(x)
    
    def arctan(self, x: generic_complex):
        return self.ctx.atan(x)
    
    def unitroots(self, N: int):
        return self.ctx.unitroots(N)

    def cooley_tukey_fft(self, x):
        N = len(x)
        if N == 1:
            return x
    
        with self.ctx.extraprec(2):
            # use two extra bits of precision to compute the even and odd FFTs
            # so that eps' = eps/4
            even = self.cooley_tukey_fft(x[0::2])
            odd = self.cooley_tukey_fft(x[1::2])
    
            W_N = self.unitroots(N)
    
            X = [self.make_complex(0)] * N
            for k in range(N // 2):
                X[k] = even[k] + W_N[-k] * odd[k]
                X[k + N // 2] = even[k] - W_N[-k] * odd[k]
    
        return X
    
    def fft(self, x: list[generic_complex], normalize=False):
        N = len(x)
        M = next_power_of_two(N)

        if M > N:
            x = x + [self.make_complex(0)] * (M - N)

        dft = self.cooley_tukey_fft(x)
        if normalize:
            N = len(dft)
            return [y / N for y in dft]
        else:
            return dft
    
    def ifft(self, x: list[generic_complex], normalize=True):
        return [self.conj(y) for y in self.fft([self.conj(xi) for xi in x], normalize)]
    
    def matrix(self, x: list):
        return self.ctx.matrix(x)
    
    def to_list(self, x):
        return x.tolist()
    
    def transpose(self, x):
        return x.transpose()
    
    def conj_transpose(self, x):
        return x.transpose_conj()
    
    def zeros(self, m: int, n: int):
        return self.ctx.zeros(m, n)
    
    def eye(self, n: int):
        return self.ctx.eye(n)
    
    def solve_system(self, A, b):
        return self.ctx.lu_solve(A, b)
    
    def qr_decomp(self, A):
        if A.rows >= A.cols:
            return self.ctx.qr(A)
        
        A1 = A[:A.rows, :A.rows]
        Q, R1 = self.ctx.qr(A1)
        R2 = Q.H * A[:A.rows, A.rows:]

        return Q, self.ctx.matrix([
            [R1[k, j] for j in range(R1.cols)] + [R2[k, j] for j in range(R2.cols)] for k in range(R1.rows)
        ])



from .. import numerics as bd

from ..nlft import NonLinearFourierSequence
from ..poly import Polynomial

def nlfft_recurse(a_star: Polynomial, b: Polynomial) -> tuple[NonLinearFourierSequence, Polynomial, Polynomial]:
    n = a_star.effective_degree() + 1

    if n == 0:
        return NonLinearFourierSequence([0]), Polynomial([0]), Polynomial([1])
    
    if n == 1:
        F0 = b[0]/a_star[0]
        r = bd.sqrt(1 + bd.abs2(F0))
        return NonLinearFourierSequence([F0]), Polynomial([F0/r]), Polynomial([1/r])
    
    m = -(-n//2) # ceil(n/2)

    Fup, xi_m, eta_m = nlfft_recurse(a_star.truncate(0,m-1), b.truncate(0,m-1))

    am_star = eta_m.conjugate() * a_star + xi_m.conjugate() * b
    bm = eta_m * b - xi_m * a_star
    bm.support_start -= m
    # Multiplies by z^(-m) (using the shift() method would duplicate the Polynomial object unnecessarily)

    Fdown, xi_mn, eta_mn = nlfft_recurse(am_star.truncate(0, n-m-1), bm.truncate(0, n-m-1))

    xi_n = eta_m.sharp() * xi_mn + xi_m * eta_mn
    eta_n = eta_m * eta_mn - xi_m.sharp() * xi_mn

    return NonLinearFourierSequence(Fup.coeffs + Fdown.coeffs), xi_n, eta_n

def inlft(a: Polynomial, b: Polynomial):
    """Computes the Inverse Non-Linear Fourier Transform using the Non-Linear Fast Fourier Transform algorithm (arXiv:2505.12615).

    Args:
        a, b (Polynomial): The pair `(a, b)` is the NLFT we want to compute the sequence for.

    Note:
        `a` must be outer. To generate an outer complementary polynomial, you can use `weiss.complete`.

    Returns:
        NonLinearFourierSequence: A sequence whose NLFT is equal to `(a, b)` (up to working precision).
    """
    if len(a.support()) != len(b.support()) or a.support().stop != 1:
        return ValueError("(a, b) must be in the image of the NLFT.")
    
    sup_start = b.support_start
    b = b.shift(-sup_start)

    n = len(a.support())
    F, _, _ = nlfft_recurse(a.conjugate(), b)
    return NonLinearFourierSequence(F.coeffs, support_start=sup_start)



from numbers import Number
from .nlft import NonLinearFourierSequence
from . import numerics as bd

from .poly_md import ComplexL0SequenceMD, PolynomialMD, to_poly_md

# Example:
# [[a, b], [c], [d, e, f], [g]]
#   fg
#   e
# bcd
# a

class StairlikeSequence2D:
    """
    Class representing a compactly supported stairlike sequence in the Z^2 grid.
    The class provides methods to compute the Non-Linear Fourier Transform (NLFT) associated with the sequence.
    """
    def __init__(self, coeffs: list, support_start: tuple[int]=(0,0)):
        if len(support_start) != 2:
            raise ValueError("support_start must be two-dimensional.")

        if not isinstance(coeffs, list):
            raise ValueError("Coefficient list must be of type list.")
        
        self.support_start = support_start

        self.coeffs = []
        self.m = [self.support_start[1]]
        for col in coeffs:
            if not all(isinstance(c, Number) for c in col):
                raise ValueError("coeffs must be a two-dimensional sequence of numbers.")
            
            if len(col) == 0:
                col = [0]
            
            self.coeffs.append([bd.make_complex(c) for c in col])
            self.m.append(self.m[-1] + len(col) - 1)

    def support_x(self):
        return range(self.support_start[0], self.support_start[0] + len(self.coeffs))
    
    def support_y(self):
        return range(self.m[0], self.m[-1] + 1)
    
    def __getitem__(self, k):
        if not (isinstance(k, tuple) and len(k) == 2):
            raise ValueError("Element index must be given as two elements.")
        
        x, y = k
        if x in self.support_x() and (self.m[x] <= y and y <= self.m[x+1]):
            return self.coeffs[x][y - self.m[x]]
        
        return bd.make_complex(0)

    def transform_bounds(self, inf, sup) -> tuple[PolynomialMD, PolynomialMD]:
        """
        Computes the Non-Linear Fourier Transform SU(2) for the subsequence within the specified range.

        Args:
            inf (int): The lower bound (included) index of the sequence for the transformation.
            sup (int): The upper bound (excluded) index of the sequence for the transformation.

        Returns:
            tuple[Polynomial, Polynomial]: The SU(2)-NLFT of the subsequence in [inf, sup].
        """
        if sup - inf <= 0:
            return PolynomialMD([[bd.make_complex(1)]], support_start=(0,0)), PolynomialMD([bd.make_complex(0)], support_start=(0,0))

        if sup - inf <= 1:
            k = inf - self.support_start[0]
            unlft = NonLinearFourierSequence(self.coeffs[k], support_start=self.m[k])
            Gk, Fk = unlft.transform()

            return to_poly_md(Gk, 2), to_poly_md(Fk, 2).shift(inf, 0) # (G_k, F_k z^k)
        
        mid = (sup + inf) // 2
        a1, b1 = self.transform_bounds(inf, mid)
        a2, b2 = self.transform_bounds(mid, sup)

        return a1 * a2 - b1 * b2.conjugate(), a1 * b2 + b1 * a2.conjugate()

    def transform(self) -> tuple[PolynomialMD, PolynomialMD]:
        """
        Computes the Non-Linear Fourier Transform (NLFT) over SU(2) associated with this sequence.

        Returns:
            tuple[Polynomial, Polynomial]: The SU(2)-NLFT of the sequence.
        """
        n = len(self.coeffs) - 1
        a, b = self.transform_bounds(self.support_start[0], self.support_start[0] + len(self.coeffs))

        if n < 0:
            return a, b
        
        b_sup_x = self.support_x()
        b_sup_y = self.support_y()
        
        b_sup = (b_sup_x, b_sup_y)
        a_sup = (range(b_sup_x.start - b_sup_x.stop, 0), range(b_sup_y.start - b_sup_y.stop, 0)) # 1s excluded

        return a.truncate(a_sup), b.truncate(b_sup)






from . import numerics as bd

from .poly import ComplexL0Sequence, Polynomial


class NonLinearFourierSequence(ComplexL0Sequence):
    """
    Class representing a finitely supported sequence of complex numbers over Z.
    The class provides methods to compute the Non-Linear Fourier Transform (NLFT) associated with the sequence.
    """

    def __init__(self, coeffs=[], support_start=0):
        """
        Initializes a non-linear Fourier sequence with a given list of complex values and support starting index.

        Args:
            coeffs (list of complex): A list of complex numbers representing the sequence. The list includes both the lower 
                                      and upper bounds of the sequence.
            support_start (int): The index of the first element of the sequence in Z. The support of the sequence will 
                                 be in the range [support_start, support_start + len(coeffs)].
        """
        super().__init__(coeffs, support_start)
    
    def transform_bounds(self, inf, sup) -> tuple[Polynomial, Polynomial]:
        """
        Computes the Non-Linear Fourier Transform SU(2) for the subsequence within the specified range.

        Args:
            inf (int): The lower bound (included) index of the sequence for the transformation.
            sup (int): The upper bound (excluded) index of the sequence for the transformation.

        Returns:
            tuple[Polynomial, Polynomial]: The SU(2)-NLFT of the subsequence in [inf, sup].
        """
        if sup - inf <= 0:
            return Polynomial([bd.make_complex(1)]), Polynomial([bd.make_complex(0)])

        if sup - inf <= 1:
            F = bd.make_complex(self[inf])
            den = bd.sqrt(1 + bd.abs2(F))
            return Polynomial([1/den]), Polynomial([F/den], inf)  # (1/den, F/den z^inf)
        
        mid = (sup + inf) // 2
        a1, b1 = self.transform_bounds(inf, mid)
        a2, b2 = self.transform_bounds(mid, sup)

        return a1 * a2 - b1 * b2.conjugate(), a1 * b2 + b1 * a2.conjugate()

    def transform(self) -> tuple[Polynomial, Polynomial]:
        """
        Computes the Non-Linear Fourier Transform (NLFT) over SU(2) associated with this sequence.

        Returns:
            tuple[Polynomial, Polynomial]: The SU(2)-NLFT of the sequence.
        """
        n = len(self.coeffs) - 1
        a, b = self.transform_bounds(self.support_start, self.support_start + n + 1)

        if n < 0:
            return a, b

        return a.truncate(-n, 0), b.truncate(self.support_start, self.support_start + n)






import numpy as np
import scipy.linalg


from .backend import NumericBackend
from .backend import generic_complex, generic_real

from ..util import next_power_of_two

def select_largest_dtypes():
    """Select the largest dtype available in the platform."""
    ctypes = ['complex512', 'complex256', 'complex192', 'complex160', 'complex128']
    ftypes = ['float256', 'float128', 'float96', 'float80', 'float64']

    for ct, ft in zip(ctypes, ftypes):
        if hasattr(np, ct):
            return getattr(np, ct), getattr(np, ft)
    
    return complex, float


class NumpyBackend(NumericBackend):
    """Numeric backend for numpy.
    
    Note:
        Using numpy allows to take advantage to optimized arithmetic computation,
        but it gives poor precision.
    """

    def __init__(self, verbose=False):
        """Initializes a numpy backend.
        """
        dtype, ftype = select_largest_dtypes()
        
        if verbose:
            print('NumpyBackend -- chosen dtypes: %s, %s' % (dtype.__name__, ftype.__name__))

        self.dtype = dtype
        self.ftype = ftype

    def __getattr__(self, item): # redirect any other function to numpy
        return getattr(np, item)

    def pi(self):
        return np.pi

    def machine_eps(self):
        return np.finfo(self.dtype).eps
    
    def machine_threshold(self):
        return 1e-8
    
    def chop(self, x: generic_complex):
        thr = self.machine_threshold()

        if np.abs(x.real) < thr:
            x.real = 0

        if np.abs(x.imag) < thr:
            x.imag = 0

        return x

    def make_complex(self, x: generic_complex):
        return self.dtype(x)

    def make_float(self, x: generic_real):
        return self.ftype(x)

    def abs2(self, x: generic_complex):
        return np.real(x) ** 2 + np.imag(x) ** 2
    
    def re(self, x: generic_complex):
        return np.real(x)
    
    def im(self, x: generic_complex):
        return np.imag(x)
    
    def arg(self, x: generic_complex):
        return np.angle(x) # TODO this only works with float64
    
    def unitroots(self, N: int):
        return [np.exp(2j*np.pi*k/N) for k in range(N)]

    def fft(self, x: list[generic_complex], normalize=False):
        N = len(x)
        M = next_power_of_two(N)

        if M > N:
            x = x + [self.make_complex(0)] * (M - N)

        if normalize:
            norm = 'forward'
        else:
            norm = 'backward'

        return np.fft.fft(np.array(x), norm=norm).tolist()
    
    def ifft(self, x: list[generic_complex], normalize=True):
        N = len(x)
        M = next_power_of_two(N)

        if M > N:
            x = x + [self.make_complex(0)] * (M - N)

        if normalize:
            norm = 'backward'
        else:
            norm = 'forward'

        return np.fft.ifft(np.array(x), norm=norm).tolist()
    
    def fft_md(self, x: list, shape: tuple=None, normalize=False):
        if normalize:
            norm = 'forward'
        else:
            norm = 'backward'

        return np.fft.fftn(np.array(x), shape, norm=norm).tolist()
    
    def ifft_md(self, x: list, shape: tuple=None, normalize=True):
        if normalize:
            norm = 'backward'
        else:
            norm = 'forward'

        return np.fft.ifftn(np.array(x), shape, norm=norm).tolist()
    
    def matrix(self, x: list):
        return np.array(x, dtype=self.dtype)
    
    def to_list(self, x):
        return x.tolist()
    
    def transpose(self, x):
        return np.transpose(x)
    
    def conj_transpose(self, x):
        return np.transpose(np.conjugate(x))
    
    def zeros(self, m: int, n: int):
        return np.zeros(shape=(m, n), dtype=self.dtype)
    
    def eye(self, n: int):
        return np.eye(n, dtype=self.dtype)
    
    def solve_system(self, A, b):
        return np.linalg.solve(A, b)
    
    def qr_decomp(self, A):
        return scipy.linalg.qr(A)
    
    def poly2cheb(self, p):
        return np.polynomial.chebyshev.poly2cheb(p).tolist()
    
    def cheb2poly(self, p):
        return np.polynomial.chebyshev.cheb2poly(p).tolist()



import unittest

import numpy as np
import scipy as sp

import nlft_qsp.numerics as bd
from nlft_qsp.rand import random_list, random_polynomial

from nlft_qsp.solvers import weiss
from nlft_qsp.solvers import riemann_hilbert
from nlft_qsp.solvers import nlfft
from nlft_qsp.solvers import layer_stripping
from nlft_qsp.solvers import half_cholesky


class RHWTestCase(unittest.TestCase):
    
    @bd.workdps(30)
    def test_rhw(self):
        b = random_polynomial(16, eta=0.5)
        a, c = weiss.ratio(b)
        
        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=bd.machine_threshold())

        self.assertAlmostEqual(max([bd.abs(c(z) - b(z)/a(z)) for z in bd.unitroots(512)]), 0, delta=bd.machine_threshold())
        self.assertEqual(c.support().stop, b.support().stop)

        Ap, Bp = riemann_hilbert.factorize(c, 10, normalize=True)
        self.assertAlmostEqual((Ap * Ap.conjugate() + Bp * Bp.conjugate() - 1).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_inlft_rhw(self):
        b = random_polynomial(16, eta=0.5)
        a, c = weiss.ratio(b)

        nlft = riemann_hilbert.inlft(b, c)
        a2, b2 = nlft.transform()

        self.assertAlmostEqual((a - a2).l2_norm(), 0, delta=100 * bd.machine_eps())
        self.assertAlmostEqual((b - b2).l2_norm(), 0, delta=100 * bd.machine_eps())

    @bd.workdps(30)
    def test_half_cholesky_ldl(self):
        c = random_polynomial(16, eta=1)
        e0 = [1] + [0] * 15

        L = half_cholesky.half_cholesky_ldl(e0, reversed(c.coeffs))
        L = np.array(bd.to_list(L), dtype=np.complex128)
        # we force numpy/scipy for this test, but L is computed with bd.

        B = np.array(bd.to_list(riemann_hilbert.toeplitz(c, 0)), dtype=np.complex128)
        K = np.eye(16) + B @ np.transpose(np.conjugate(B))
        for k in range(16): # just to suppress the annoying ComplexWarning
            K[k, k] = np.real(K[k, k])

        L2, _, perm = sp.linalg.ldl(K)
        L2 = L2[perm, :]

        self.assertAlmostEqual(np.linalg.norm(L - L2), 0, delta=1e-5)

    @bd.workdps(30)
    def test_inlft_rhw_hc(self):
        b = random_polynomial(16, eta=0.5)
        a, c = weiss.ratio(b)

        nlft = half_cholesky.inlft(b, c)
        a2, b2 = nlft.transform()

        self.assertAlmostEqual((a - a2).l2_norm(), 0, delta=100 * bd.machine_eps())
        self.assertAlmostEqual((b - b2).l2_norm(), 0, delta=100 * bd.machine_eps())

    @bd.workdps(30)
    def test_inlft_nlfft(self):
        b = random_polynomial(1600, eta=0.5)
        a = weiss.complete(b)

        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=bd.machine_threshold())

        nlft = nlfft.inlft(a, b)
        a2, b2 = nlft.transform()

        self.assertAlmostEqual((a - a2).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((b - b2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_inlft_layer_stripping(self):
        b = random_polynomial(1600, eta=0.5)
        a = weiss.complete(b)

        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=bd.machine_threshold())

        nlft = layer_stripping.inlft(a, b)
        a2, b2 = nlft.transform()

        self.assertAlmostEqual((a - a2).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((b - b2).l2_norm(), 0, delta=bd.machine_threshold())


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

import nlft_qsp.numerics as bd

from nlft_qsp.poly import ChebyshevTExpansion, Polynomial
from nlft_qsp.nlft import NonLinearFourierSequence
from nlft_qsp.qsp import ChebyshevQSPPhaseFactors, GQSPPhaseFactors, QSVTPhaseFactors, XQSPPhaseFactors, YQSPPhaseFactors, gqsp_solve, chebqsp_solve, xqsp_solve, yqsp_solve, qsvt_solve
from nlft_qsp.rand import random_polynomial, random_real_polynomial, random_real_sequence, random_sequence


class QSPTestCase(unittest.TestCase):

    @bd.workdps(30)
    def test_xqsp_phase_factors(self):
        qsp = XQSPPhaseFactors([bd.pi()/3, bd.pi()/6, bd.pi()/4])

        P, Q = qsp.polynomials()
        for k, c in enumerate([-0.53033, -0.482963, 0.306186]):
            self.assertAlmostEqual(P[k], c, delta=10e-7)

        for k, c in enumerate([0.53033j, -0.12941j, 0.306186j]):
            self.assertAlmostEqual(Q[k], c, delta=10e-7)

        F = NonLinearFourierSequence([1j*c for c in random_real_sequence(1000, 10)])
        qsp = XQSPPhaseFactors.from_nlfs(F)
        F2 = qsp.to_nlfs()

        self.assertAlmostEqual((F - F2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_yqsp_phase_factors(self):
        qsp = YQSPPhaseFactors([bd.pi()/3, bd.pi()/6, bd.pi()/4])

        P, Q = qsp.polynomials()
        for k, c in enumerate([-0.53033, -0.482963, 0.306186]):
            self.assertAlmostEqual(P[k], c, delta=10e-7)

        for k, c in enumerate([0.53033, -0.12941, 0.306186]):
            self.assertAlmostEqual(Q[k], c, delta=10e-7)
            
        F = NonLinearFourierSequence(random_real_sequence(1000, 10))
        qsp = YQSPPhaseFactors.from_nlfs(F)
        F2 = qsp.to_nlfs()

        self.assertAlmostEqual((F - F2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_gqsp_phase_factors(self):
        F = NonLinearFourierSequence(random_sequence(1000, 10))
        qsp = GQSPPhaseFactors.from_nlfs(F)
        F2 = qsp.to_nlfs()

        self.assertAlmostEqual((F - F2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_chebqsp_phase_factors(self):
        qsp = ChebyshevQSPPhaseFactors([bd.pi()/3, bd.pi()/6, bd.pi()/4])

        P, Q = qsp.polynomials(mode='laurent')
        for k, c in enumerate([-0.112072+0.418258j, 0, -0.482963-0.12941j, 0, -0.112072+0.418258j]):
            self.assertAlmostEqual(P[k - 2], c, delta=10e-7)

        for k, c in enumerate([-0.418258-0.112072j, 0, 0, 0, 0.418258+0.112072j]):
            self.assertAlmostEqual(Q[k - 2], c, delta=10e-7)

    @bd.workdps(30)
    def test_gqsp_polynomials(self):
        nlft = NonLinearFourierSequence(random_sequence(1000000, 16))
        qsp = GQSPPhaseFactors.from_nlfs(nlft)

        a, b = nlft.transform()
        P, Q = qsp.polynomials()

        self.assertAlmostEqual((a.shift(15) - P).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((b - Q).l2_norm(), 0, delta=bd.machine_threshold())

        qsp2 = GQSPPhaseFactors.from_nlfs(nlft, alpha=1)
        P2, Q2 = qsp2.polynomials()
        self.assertAlmostEqual((P2 - P * bd.exp(1j)).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((Q2 - Q).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_qsp_polynomial_laurent_analytic(self):
        nlft = NonLinearFourierSequence(random_sequence(100, 17))
        qsp = GQSPPhaseFactors.from_nlfs(nlft)

        P, Q = qsp.polynomials()
        Pl, Ql = qsp.polynomials(mode='laurent')

        self.assertEqual(P.support_start, 0)
        self.assertEqual(Q.support_start, 0)
        n = P.effective_degree()

        self.assertEqual(Pl.support_start, -n)
        self.assertEqual(Ql.support_start, -n)
        for k in P.support():
            self.assertAlmostEqual(Pl[2*k - n], P[k], delta=bd.machine_threshold())
            self.assertAlmostEqual(Ql[2*k - n], Q[k], delta=bd.machine_threshold())
            self.assertAlmostEqual(Pl[2*k - n + 1], 0, delta=bd.machine_threshold())
            self.assertAlmostEqual(Ql[2*k - n + 1], 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_gqsp_solver(self):
        P = random_polynomial(16, eta=0.5)

        qsp = gqsp_solve(P, mode='nlft')
        Q2, P2 = qsp.polynomials()

        self.assertAlmostEqual((P - P2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_gqsp_solver_ix(self):
        P = random_polynomial(1024, eta=0.5)

        qsp1 = gqsp_solve(P, mode='nlft')
        qsp2 = gqsp_solve(P)
        Q1, P1 = qsp1.polynomials() # (Q, P)
        P2, Q2 = qsp2.polynomials() # (P, iQ)

        self.assertAlmostEqual((P - P2).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((P1 - P2).l2_norm(), 0, delta=bd.machine_threshold())

        self.assertAlmostEqual((1j*Q1 - Q2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_xqsp_solver(self):
        P = random_real_polynomial(16, eta=0.5)

        qsp = xqsp_solve(1j*P, mode='nlft')
        Q2, P2 = qsp.polynomials()
        self.assertAlmostEqual((1j*P - P2).l2_norm(), 0, delta=bd.machine_threshold())

        qsp = xqsp_solve(P)
        P2, Q2 = qsp.polynomials()
        self.assertAlmostEqual((P - P2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_yqsp_solver(self):
        P = random_real_polynomial(16, eta=0.5)

        qsp = yqsp_solve(P, mode='nlft')
        Q2, P2 = qsp.polynomials()
        self.assertAlmostEqual((P - P2).l2_norm(), 0, delta=bd.machine_threshold())

        qsp = yqsp_solve(P)
        P2, Q2 = qsp.polynomials()
        self.assertAlmostEqual((P - P2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_chebqsp_solve(self):
        coef_odd = [0, 0.1, 0, -0.3, 0, 0.2, 0, 0.14]
        coef_even = [0.3, 0, -0.2, 0, 0.1, 0, 0.19]
        for coef in (coef_odd, coef_even):
            T = ChebyshevTExpansion(coef)
            phase_factors = chebqsp_solve(T)
            P, Q = phase_factors.polynomials(mode="laurent")

            # Degree
            self.assertEqual(P.effective_degree(), T.degree()*2)
            
            # Phase factor symmetry
            phi = list(phase_factors.phi)
            phi[-1] -= bd.pi()/2
            for p1, p2 in zip(phi, phi[::-1]):
                self.assertAlmostEqual(p1, p2, delta=bd.machine_threshold())

            # Polynomial relationships
            for alpha in np.linspace(0, 2*np.pi, 100):
                z = bd.exp(1j*alpha)
                x = bd.cos(alpha)
                self.assertAlmostEqual(bd.re(P(z)), T(x), delta=bd.machine_threshold())

    def test_qsvt_polynomials(self):
        phi = [bd.pi()/3, bd.pi()/6, bd.pi()/4]

        #Expected polynomial = e^(5 i \pi/12) + e^(i \pi/4) (i - e^(i \pi/6)) x^2
        Texp = ChebyshevTExpansion.from_polynomial(Polynomial([bd.exp(1j*5*bd.pi()/12), 0, bd.exp(1j*bd.pi()/4)*(1j - bd.exp(1j*bd.pi()/6))]))

        qsp = QSVTPhaseFactors(phi)
        P, _ = qsp.polynomials(mode='laurent')
        T = ChebyshevTExpansion.from_laurent_polynomial(P)

        for k in range(T.degree() + 1):
            self.assertAlmostEqual(T[k], Texp[k], delta=bd.machine_threshold())

    def test_qsvt_solve(self):
        coef_odd = [0, 0.1, 0, -0.3, 0, 0.2, 0, 0.14]
        coef_even = [0.3, 0, -0.2, 0, 0.1, 0, 0.19]
        for coef in (coef_odd, coef_even):
            T = ChebyshevTExpansion(coef)
            phase_factors = qsvt_solve(T)
            P, Q = phase_factors.polynomials(mode="laurent")

            # Degree
            self.assertEqual(P.effective_degree(), T.degree()*2)
            
            # Polynomial relationships
            for alpha in np.linspace(0, 2*np.pi, 100):
                z = bd.exp(1j*alpha)
                x = bd.cos(alpha)
                self.assertAlmostEqual(bd.re(P(z)), T(x), delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_chebqsp_to_qsvt(self):
        seq = random_real_sequence(2*bd.pi(), 12)
        qsp = ChebyshevQSPPhaseFactors(seq)
        qsvt = QSVTPhaseFactors.from_chebqsp(qsp)
        
        P1, Q1 = qsp.polynomials(mode='laurent')
        P2, Q2 = qsvt.polynomials(mode='laurent')

        self.assertAlmostEqual((P1 - P2).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((Q1 - Q2).l2_norm(), 0, delta=bd.machine_threshold())

    @bd.workdps(30)
    def test_qsvt_to_chebqsp(self):
        seq = random_real_sequence(2*bd.pi(), 12)
        qsvt = QSVTPhaseFactors(seq)
        qsp = qsvt.to_chebqsp()
        
        P1, Q1 = qsp.polynomials(mode='laurent')
        P2, Q2 = qsvt.polynomials(mode='laurent')

        self.assertAlmostEqual((P1 - P2).l2_norm(), 0, delta=bd.machine_threshold())
        self.assertAlmostEqual((Q1 - Q2).l2_norm(), 0, delta=bd.machine_threshold())

if __name__ == '__main__':
    unittest.main()



from . import numerics as bd

from .poly import ChebyshevTExpansion, Polynomial
from .nlft import NonLinearFourierSequence
from .numerics.backend import generic_real, generic_complex
from .approximate import chebyshev_approximate

from .solvers import nlfft, weiss

def is_definite_parity(P: Polynomial, n: int = -1) -> bool:
    """Returns whether the polynomial has the parity of n. If n is not defined, then n = index of last coefficient of P."""
    if n < 0:
        n = P.support().stop - 1

    for k in P.support():
        if (k - n) % 2 != 0 and abs(P[k]) > bd.machine_threshold():
            return False

    return True


def analytic_to_laurent(P: Polynomial, n: int = -1) -> Polynomial:
    """Converts the given polynomial `p[0] + p[1] z + ... + p[n] z^n` into a definite-parity Laurent polynomial `p[0] z^(-n) + p[1] z^(-n+2) + ... + p[n] z^n`. If n is not given, then the degree of P will be considered.
    
    Note:
        No check is done on whether P is actually analytic, any negative-degree coefficients are not considered."""
    if n < 0:
        n = P.support().stop - 1

    Pl = Polynomial([0], support_start=-n)
    for k in range(n+1):
        Pl[2*k - n] = P[k]
    
    return Pl

def laurent_to_analytic(P: Polynomial, n: int = -1) -> Polynomial:
    """Converts the given definite-parity Laurent polynomial `p[0] z^(-n) + p[1] z^(-n+2) + ... + p[n] z^n` into an analytic polynomial `p[0] + p[1] z + ... + p[n] z^n`. If n is not given, then n = the Laurent degree of P.
    
    Note:
        No check is done on whether P is of definite-parity, and the other coefficients are not considered.
    """
    if n < 0:
        n = max(abs(P.support().start), abs(P.support().stop - 1))
    
    return Polynomial([P[2*k - n] for k in range(n+1)])

def chebyshev_to_laurent(c: list[generic_complex]) -> Polynomial:
    """Returns the Laurent polynomial equivalent to the Chebyshev expansion."""
    P = Polynomial(c)
    return (P + P.conjugate())/2


def phase_prefactor(F: generic_complex) -> generic_real:
    """Computes the phase prefactor for the Fourier sequence coefficient F."""

    if bd.abs(F) < bd.machine_threshold():
        return 0
    
    if bd.abs(bd.im(F)) < bd.machine_threshold():
        return -(bd.pi()/4)

    return -bd.arctan(bd.re(F)/bd.im(F))/2

class PhaseFactors:
    """Set of phase factors for a general Quantum Signal Processing protocol.
    It also provides methods to construct polynomials generated by QSP protocols.
    Each subclass of this class represents a different QSP ansatz."""
    def duplicate(self):
        raise NotImplementedError()

    def processing_operator(self, k: int):
        """Returns the k-th signal processing operator according to the given QSP variant."""
        raise NotImplementedError()
    
    def signal_operator(self, P1: Polynomial, Q1: Polynomial, P2: Polynomial, Q2: Polynomial) -> tuple[Polynomial, Polynomial]:
        """Returns the pair of polynomials given by multiplying `(P1, Q1) W(z) (P2, Q2)`, where `W(z)` is the signal operator.
        
        Note:
            This might not reflect the signal operator as expressed in the original papers. Some basis transformations are implicitly made for computational efficiency."""
        # W(z) = diag(z, z^(-1))
        zP1 = P1.shift(1)  # z*P1
        zQ1 = Q1.shift(-1) # z^(-1)*Q1
        return zP1 * P2 - zQ1 * Q2.conjugate(), zP1 * Q2 + zQ1 * P2.conjugate()
    
    def protocol_conjugation(self, P, Q):
        """Given (P, Q), this applies a final conjugation to the whole protocol, e.g., with the Hadamard gate."""
        return P, Q
    
    def processing_operator_conjugation(self, a, b):
        """This applies a conjugation to each signal processing operator returned by `processing_operator()`, e.g., with the Hadamard gate.
        This is done mainly to make `processing_operator()` return the operators as expressed by the original ansatze in the papers, while keeping computational efficiency and numerical stability."""
        return a, b
    
    def degree(self):
        """Returns the degree of the polynomials generated by the QSP protocol."""
        raise NotImplementedError()
    
    def iX(self):
        """Returns a new QSP protocol, obtained by multiplying the given QSP protocol by iX on the right, where X is the Pauli matrix.
        This will make the generated polynomials undergo the transformation (P, Q) -> (iQ, iP).
        This method is useful to bring to swap the places of the two polynomials, to switch between NLFT and QSP conventions.
        
        Raises:
            ValueError: if multiplying by iX does not preserve the subalgebra of the phase factors."""
        raise ValueError("Multiplying by iX is not possible.")
    
    def iY(self):
        """Returns a new QSP protocol, obtained by multiplying the given QSP protocol by iY on the right, where Y is the Pauli matrix.
        This will make the generated polynomials undergo the transformation (P, Q) -> (-Q, P).
        This method is useful to bring to swap the places of the two polynomials, to switch between NLFT and QSP conventions.
        
        Raises:
            ValueError: if multiplying by iY does not preserve the subalgebra of the phase factors."""
        raise ValueError("Multiplying by iY is not possible.")
    
    def iZ(self):
        """Returns a new QSP protocol, obtained by multiplying the given QSP protocol by iZ on the right, where Z is the Pauli matrix.
        This will make the generated polynomials undergo the transformation (P, Q) -> (iP, -iQ).
        This method is useful to bring to swap the places of the two polynomials, to switch between NLFT and QSP conventions.
        
        Raises:
            ValueError: if multiplying by iZ does not preserve the subalgebra of the phase factors."""
        raise ValueError("Multiplying by iZ is not possible.")
    
    def polynomials_bounds(self, inf: int, sup: int) -> tuple[Polynomial, Polynomial]:
        """Returns the pair of polynomials `(P, Q) = A[inf] w A[inf+1] w ... w A[sup-1] w A[sup]`,
        where w is the signal operator.
        
        Note:
            We use the Laurent picture to compute, so that we can use `Polynomial.conjugate()`."""
        if sup - inf < 0:
            return Polynomial([bd.make_complex(1)]), Polynomial([bd.make_complex(0)])
        if sup - inf <= 0:
            p, q = self.processing_operator(inf)
            p, q = self.processing_operator_conjugation(p, q)
            return Polynomial([p]), Polynomial([q])
        
        mid = (sup + inf) // 2
        P1, Q1 = self.polynomials_bounds(inf, mid)
        P2, Q2 = self.polynomials_bounds(mid+1, sup)

        return self.signal_operator(P1, Q1, P2, Q2) # (P1, Q1) W(z) (P2, Q2)

    def polynomials(self, inf: int = 0, sup: int = -1, mode: str = 'analytic') -> tuple[Polynomial, Polynomial]:
        """Returns the pair of polynomials (P, Q) generated by the given set of phase factors.
        The polynomials are computed with a divide and conquer strategy (see arXiv:2410.06409).
        
        Args:
            mode (str): Either 'analytic' or 'laurent', indicating whether an analytic or a Laurent QSP protocol should be composed.
        """
        if sup < 0:
            sup = self.degree()

        Pl, Ql = self.polynomials_bounds(inf, sup)
        Pl, Ql = self.protocol_conjugation(Pl, Ql)
        match mode:
            case 'analytic': # convert from Laurent to analytic picture
                return laurent_to_analytic(Pl), laurent_to_analytic(Ql)
            case 'laurent':
                return Pl, Ql
            case _:
                raise ValueError("mode can only be 'analytic' or 'laurent'.")
    
    def to_nlfs(self) -> NonLinearFourierSequence:
        """Returns the Non-Linear Fourier Sequence generating (z^{-n} P, Q),
        where (P, Q) is the pair of polynomial generated by the given set of GQSP phase factors.
        
        Note: if the phase factors are not canonical, then the phase of the leading coefficient of P is adjusted
        so that it becomes real and positive, and (z^{-n} P, Q) is in the image of the NLFT."""
        raise NotImplementedError()
    
    
class XQSPPhaseFactors(PhaseFactors):
    """Phase factors for a X-constrained QSP protocol.
    
    Signal operator: `W(z) = diag(z, z^(-1))`,
    
    Processing operators: `A[k] = exp(I*phi[k]*X)`."""
    def __init__(self, phi: list[generic_real]):
        self.phi = list(phi)

    def duplicate(self):
        return XQSPPhaseFactors(self.phi)
    
    def processing_operator(self, k: int): # exp(i phi[k] X)
        return bd.cos(self.phi[k]), 1j*bd.sin(self.phi[k])
    
    def degree(self):
        return len(self.phi) - 1

    def iX(self):
        pf = self.duplicate()
        pf.phi[-1] += bd.pi()/2
        return pf
    
    def to_nlfs(self) -> NonLinearFourierSequence:
        return NonLinearFourierSequence([1j*bd.tan(phik) for phik in self.phi])
    
    @classmethod
    def from_nlfs(cls, F: NonLinearFourierSequence) -> PhaseFactors:
        """Computes the XQSP phase factors for a given imaginary NLFT sequence.
        If `NLFT(F) = (a, b)`, then the returned phase factors will implement `(z^n a, b)` in the analytic picture.
    
        Args:
            F (NonLinearFourierSequence): The imaginary sequence to be converted to phase factors.

        Raises:
            ValueError if F is not imaginary.
    
        Note:
            The support start of F is ignored, so the support of b is assumed to start at 0."""
        if not F.is_imaginary():
            raise ValueError("The Non-Linear Fourier sequence must be imaginary in order to be turned into a XQSP protocol.")

        return XQSPPhaseFactors([bd.arctan(bd.im(Fk)) for Fk in F.coeffs])
    
class YQSPPhaseFactors(PhaseFactors):
    """Phase factors for a Y-constrained QSP protocol.

    Signal operator: `W(z) = diag(z, z^(-1))`,
    
    Processing operators: `A[k] = exp(I*phi[k]*Y)`."""
    def __init__(self, phi: list[generic_real]):
        self.phi = list(phi)

    def duplicate(self):
        return YQSPPhaseFactors(self.phi)
    
    def processing_operator(self, k: int): # exp(i phi[k] Y)
        return bd.cos(self.phi[k]), bd.sin(self.phi[k])
    
    def degree(self):
        return len(self.phi) - 1

    def iY(self):
        pf = self.duplicate()
        pf.phi[-1] += bd.pi()/2
        return pf
    
    def to_nlfs(self) -> NonLinearFourierSequence:
        return NonLinearFourierSequence([bd.tan(phik) for phik in self.phi])
    
    @classmethod
    def from_nlfs(cls, F: NonLinearFourierSequence) -> PhaseFactors:
        """Computes the YQSP phase factors for a given real NLFT sequence.
        If `NLFT(F) = (a, b)`, then the returned phase factors will implement `(z^n a, b)` in the analytic picture.
    
        Args:
            F (NonLinearFourierSequence): The real sequence to be converted to phase factors.

        Raises:
            ValueError if F is not real.
    
        Note:
            The support start of F is ignored, so the support of b is assumed to start at 0."""
        if not F.is_real():
            raise ValueError("The Non-Linear Fourier sequence must be real in order to be turned into a YQSP protocol.")

        return YQSPPhaseFactors([bd.arctan(bd.re(Fk)) for Fk in F.coeffs])

class GQSPPhaseFactors(PhaseFactors):
    """Phase factors for a GQSP protocol.
    
    Signal operator: `W(z) = diag(z, z^(-1))`,
    
    Processing operators: `A[0] = exp(I*lbd*Z) exp(I*phi[0]*X) exp(I*theta[0]*Z), A[k] = exp(I*phi[k]*X) exp(I*theta[k]*Z)`.
    
    Note:
        This class follows the convention of arXiv:2503.03026, Theorem 2, which is different from the original GQSP convention. If the convention of arXiv:2308.01501 Theorem 3 is desired, then one should use the `to_mw_gqsp()` method."""
    def __init__(self, phi: list[generic_real], lbd: generic_real=0, theta: list[generic_real]=None):
        self.lbd = lbd
        self.phi = list(phi)

        if theta is None:
            theta = [0] * len(phi)

        self.theta = list(theta)
        
        if len(theta) < len(phi):
            theta += [0] * (len(phi) - len(theta))

        if len(theta) > len(phi):
            phi += [0] * (len(theta) - len(phi))

    def duplicate(self):
        return GQSPPhaseFactors(self.phi, self.lbd, self.theta)
    
    def processing_operator(self, k: int):
        if k == 0: # exp(i lbd Z) exp(i phi X) exp(i theta Z)
            return bd.exp(1j*(self.lbd + self.theta[k]))*bd.cos(self.phi[k]), \
                1j*bd.exp(1j*(self.lbd - self.theta[k]))*bd.sin(self.phi[k])
        else: # exp(i phi X) exp(i theta Z)
            return bd.exp(1j*self.theta[k])*bd.cos(self.phi[k]), 1j*bd.exp(-1j*self.theta[k])*bd.sin(self.phi[k])
    
    def degree(self):
        return len(self.phi) - 1

    def iX(self):
        pf = self.duplicate()
        pf.theta[-1] = -pf.theta[-1]
        pf.phi[-1] += bd.pi()/2
        return pf
    
    def iY(self):
        return self.iZ().iX()
    
    def iZ(self):
        pf = self.duplicate()
        pf.theta[-1] += bd.pi()/2
        return pf
    
    def phase_offset(self) -> generic_real:
        """Returns the phase of the leading coefficient of P, where (P, Q) is the pair of polynomials generated by this set."""
        return self.lbd + sum(self.theta)
    
    def to_xqsp(self):
        """Converts the QSP phase factors into X-constrained QSP phase factors.
        
        Raises:
            ValueError: If the phase factors do not lie in the X-constrained subalgebra."""
        F = self.to_nlfs()
        if not F.is_imaginary():
            raise ValueError("The phase factors are not reducible to X-constrained QSP.")
        
        return XQSPPhaseFactors.from_nlfs(F)
    
    def to_yqsp(self):
        """Converts the QSP phase factors into Y-constrained QSP phase factors.
        
        Raises:
            ValueError: If the phase factors do not lie in the Y-constrained subalgebra."""
        F = self.to_nlfs()
        if not F.is_real():
            raise ValueError("The phase factors are not reducible to Y-constrained QSP.")
        
        return YQSPPhaseFactors.from_nlfs(F)
    
    def to_nlfs(self) -> NonLinearFourierSequence:
        n = self.degree()
        alpha = self.phase_offset() # phase of the leading coefficient of P

        psi = [bd.make_float(0)] * (n+1) # prefactors

        psi[0] = self.lbd - alpha/2
        for k in range(n):
            psi[k+1] = self.theta[k] + psi[k]

        phi = self.phi
        return NonLinearFourierSequence([1j*bd.tan(phik)*bd.exp(2j*psik) for phik, psik in zip(phi, psi)])
    
    @classmethod
    def from_nlfs(cls, F: NonLinearFourierSequence, alpha: generic_real = 0) -> PhaseFactors:
        """Computes the GQSP phase factors for a given NLFT sequence.
        If `NLFT(F) = (a, b)`, then the returned phase factors will implement `(exp(i alpha) z^n a, b)` in the analytic picture.
    
        Args:
            F (NonLinearFourierSequence): The sequence to be converted to phase factors.
            alpha (float): In the pair of polynomials (P, Q) generated by the returned phase factors,
            P will be multiplied by `exp(i alpha)`.
    
        Note:
            The support start of F is ignored, so the support of b is assumed to start at 0."""
        psi = [phase_prefactor(Fk) for Fk in F.coeffs]
        lbd = psi[0]

        phi = [bd.arctan(bd.im(Fk * bd.exp(-2j * psik))) for Fk, psik in zip(F, psi)]

        psi += [bd.make_float(0)] # we add psi[n+1] to compute theta
        theta = [psi[k+1] - psi[k] for k in range(len(phi))]

        lbd += alpha/2
        theta[-1] += alpha/2
        return GQSPPhaseFactors(phi, lbd, theta)
    
    def to_mw_gqsp(self) -> tuple[list[generic_real], list[generic_real], generic_real]:
        """Converts the GQSP phase factors to the convention of arXiv:2308.01501 Theorem 3.

        In particular, a protocol should be constructed so that `A[k] = R(theta[k], phi[k], 0)` for k < n and `A[n] = R(theta[n], phi[n], lbd)` where `R(theta, phi, lbd)` is as given in arXiv:2308.01501.

        Returns:
            The phase factors theta[k], phi[k] (as lists) and lbd, in this order."""
        n = self.degree()

        alpha = - self.lbd - sum(self.theta) - (n+1) * bd.pi() # needed to counterbalance the global phase

        mw_theta = [self.phi[k] for k in range(n+1)]
        mw_phi = [2 * self.lbd - bd.pi()/2 + alpha] + [2 * self.theta[k-1] - bd.pi() for k in range(1, n+1)]
        mw_lbd = 2 * self.theta[n] - bd.pi()/2 + alpha

        return mw_theta, mw_phi, mw_lbd

    
class ChebyshevQSPPhaseFactors(XQSPPhaseFactors):
    """Phase factors for a Chebyshev QSP protocol.

    Signal operator: `W(z) = (x, I*sqrt(1 - x^2))` with x = (z + z^(-1))/2.
    
    Processing operators: `A[k] = exp(I*phi[k]*Z)`.

    Note:
        This is the ansatz of Theorem 9 arXiv:2105.02859, but the polynomial
        construction is implemented by conjugating XQSP with a Hadamard gate."""
    def processing_operator(self, k: int): # exp(i phi[k] Z)
        return bd.exp(1j*self.phi[k]), bd.make_complex(0)

    def protocol_conjugation(self, P, Q): # Return the polynomials (P', Q') = H (P, Q) H
        return ((P + P.conjugate()) + (Q - Q.conjugate()))/2, ((P - P.conjugate()) - (Q + Q.conjugate()))/2
    
    def processing_operator_conjugation(self, a, b):
        return bd.re(a) + 1j*bd.im(b), 1j*bd.im(a) - bd.re(b)

    def duplicate(self):
        return ChebyshevQSPPhaseFactors(self.phi)
    
    def iX(self):
        raise ValueError("Multiplying by iX is not possible.")
    
    def iZ(self):
        return super.iX() # applying iZ is equivalent to applying iX, by the Hadamard conjugation
    
class QSVTPhaseFactors(ChebyshevQSPPhaseFactors):
    """Phase factors for a QSVT/Reflection QSP protocol.

    Signal operator: `R(x) = (x, sqrt(1 - x^2))` where `R(x)` is Hermitian with x = (z + z^(-1))/2.
    
    Processing operators: `A[k] = exp(I*phi[k]*Z)`.

    Note:
        This is the ansatz of Corollary 8 arXiv:1806.01838, but the polynomial
        construction is implemented by implicitly adjusting the phase factors from Chebyshev QSP, see arXiv:2105.02859, (A5)."""
    def protocol_conjugation(self, P, Q):
        d = self.degree()
        return ((P + P.conjugate()) + (Q - Q.conjugate()))*((-1j)**(d+1)/2), \
               ((P - P.conjugate()) - (Q + Q.conjugate()))*((-1j)**d/2)
    
    def processing_operator_conjugation(self, a, b):
        a *= 1j # R(x) = -I*exp(I*pi*Z/4)*W(x)*exp(I*pi*Z/4)
        return bd.re(a) + 1j*bd.im(b), 1j*bd.im(a) - bd.re(b)

    def duplicate(self):
        return QSVTPhaseFactors(self.phi)

    @classmethod
    def from_chebqsp(cls, pf: ChebyshevQSPPhaseFactors):
        """Returns a set of QSVT phase factors constructing the same left chebyshev expansion as the given QSP protocol."""
        d = pf.degree()
        phi = [0] * (d+1)

        phi[0] = pf.phi[0] + (2*d - 1)*bd.pi()/4
        for k in range(1, d):
            phi[k] = pf.phi[k] - bd.pi()/2
        phi[d] = pf.phi[d] - bd.pi()/4

        return QSVTPhaseFactors(phi)

    def to_chebqsp(self) -> ChebyshevQSPPhaseFactors:
        """Returns a set of Chebyshev QSP phase factors constructing the same left chebyshev expansion as the given QSVT protocol."""
        d = self.degree()
        phi = [0] * (d+1)

        phi[0] = self.phi[0] - (2*d - 1)*bd.pi()/4
        for k in range(1, d):
            phi[k] = self.phi[k] + bd.pi()/2
        phi[d] = self.phi[d] + bd.pi()/4

        return ChebyshevQSPPhaseFactors(phi)


#### QSP SOLVERS

def __riemann_hilbert_weiss(P: Polynomial) -> NonLinearFourierSequence:
    Q = weiss.complete(P)
    return nlfft.inlft(Q, P) # NLFT(F) = (Q, P)

def gqsp_solve(P: Polynomial, mode='qsp') -> GQSPPhaseFactors:
    r"""Returns the set of phase factors for a Generalized QSP protocol producing the given polynomial.
    A complementary Q will be computed with Weiss' algorithm.

    Args:
        mode (str): Whether the phase factors should produce (P, Q) (`'qsp'`), or (Q, P) (`'nlft'`).
    
    Note:
        The sup norm of P should be bounded by :math:`1 - \eta < 1`.
        The time required by the algorithm to compute the phase factors will scale with :math:`1/\eta`.
        
        The support_start of P will be ignored."""
    if 1 - P.sup_norm(4*P.effective_degree()) < bd.machine_threshold():
        raise ValueError("The given polynomial cannot be too close to or larger than one on the unit circle.")
    
    match mode:
        case "qsp":
            P = -1j * Polynomial(P.coeffs, 0) # (Q, -iP) -> (P, iQ)
        case "nlft":
            P = Polynomial(P.coeffs, 0)
        case _:
            raise ValueError("The given mode does not exist. Only modes available are 'qsp', 'nlft'.")
    
    F = __riemann_hilbert_weiss(P) # NLFT(F) = (Q, P)

    if mode != "qsp":
        return GQSPPhaseFactors.from_nlfs(F)
    return GQSPPhaseFactors.from_nlfs(F).iX()

def xqsp_solve(P: Polynomial, mode='qsp') -> XQSPPhaseFactors:
    r"""Returns the set of phase factors for a X-constrained QSP protocol producing the given polynomial.
    A complementary Q will be computed with Weiss' algorithm.

    Args:
        mode (str): Whether the phase factors should produce (P, Q) (`'qsp'`), or (Q, P) (`'nlft'`).

    Raises:
        ValueError: If P does not lie in the X-constrained subalgebra.
    
    Note:
        The sup norm of P should be bounded by :math:`1 - \eta < 1`.
        The time required by the algorithm to compute the phase factors will scale with :math:`1/\eta`.
        
        The support_start of P will be ignored. This is a solver for analytic QSP. In order to obtain phase factors for Laurent XQSP, use `xqsp_solve_laurent`."""
    if 1 - P.sup_norm(4*P.effective_degree()) < bd.machine_threshold():
        raise ValueError("The given polynomial cannot be too close to or larger than one on the unit circle.")
    
    match mode:
        case "qsp":
            P = -1j * Polynomial(P.coeffs, 0) # (Q, -iP) -> (P, iQ)
        case "nlft":
            P = Polynomial(P.coeffs, 0)
        case _:
            raise ValueError("The given mode does not exist. Only modes available are 'qsp', 'nlft'.")
    
    F = __riemann_hilbert_weiss(P) # NLFT(F) = (Q, P)

    if mode != "qsp":
        return XQSPPhaseFactors.from_nlfs(F)
    return XQSPPhaseFactors.from_nlfs(F).iX()

def xqsp_solve_laurent(P: Polynomial, mode='qsp') -> XQSPPhaseFactors:
    r"""Returns the set of phase factors for a X-constrained QSP protocol producing the given definite-parity polynomial. A complementary Q will be computed with Weiss' algorithm.

    Args:
        mode (str): Whether the phase factors should produce (P, Q) (`'qsp'`), or (Q, P) (`'nlft'`).

    Raises:
        ValueError: If P does not lie in the X-constrained subalgebra or P has not definite-parity.
    
    Note:
        The sup norm of P should be bounded by :math:`1 - \eta < 1`.
        The time required by the algorithm to compute the phase factors will scale with :math:`1/\eta`.
        
        The support_start of P will be ignored. In order to obtain phase factors for Laurent XQSP, first convert the polynomial into analytic form."""
    if not is_definite_parity(P):
        raise ValueError("Laurent polynomial is not of definite parity.")
    
    return xqsp_solve(laurent_to_analytic(P), mode=mode)

def yqsp_solve(P: Polynomial, mode='qsp') -> YQSPPhaseFactors:
    r"""Returns the set of phase factors for a Y-constrained QSP protocol producing the given polynomial.
    A complementary Q will be computed with Weiss' algorithm.

    Args:
        mode (str): Whether the phase factors should produce (P, Q) (`'qsp'`), or (Q, P) (`'nlft'`).

    Raises:
        ValueError: If P does not lie in the Y-constrained subalgebra.
    
    Note:
        The sup norm of P should be bounded by :math:`1 - \eta < 1`.
        The time required by the algorithm to compute the phase factors will scale with :math:`1/\eta`.
        
        The support_start of P will be ignored. This is a solver for analytic QSP. In order to obtain phase factors for Laurent YQSP, use `yqsp_solve_laurent`."""
    if 1 - P.sup_norm(4*P.effective_degree()) < bd.machine_threshold():
        raise ValueError("The given polynomial cannot be too close to or larger than one on the unit circle.")
    
    match mode:
        case "qsp":
            P = -Polynomial(P.coeffs, 0) # (Q, -P) -> (P, Q)
        case "nlft":
            P = Polynomial(P.coeffs, 0)
        case _:
            raise ValueError("The given mode does not exist. Only modes available are 'qsp', 'nlft'.")
    
    F = __riemann_hilbert_weiss(P) # NLFT(F) = (Q, P)

    if mode != "qsp":
        return YQSPPhaseFactors.from_nlfs(F)
    return YQSPPhaseFactors.from_nlfs(F).iY()

def yqsp_solve_laurent(P: Polynomial, mode='qsp') -> YQSPPhaseFactors:
    r"""Returns the set of phase factors for a Y-constrained QSP protocol producing the given definite-parity polynomial. A complementary Q will be computed with Weiss' algorithm.

    Args:
        mode (str): Whether the phase factors should produce (P, Q) (`'qsp'`), or (Q, P) (`'nlft'`).

    Raises:
        ValueError: If P does not lie in the X-constrained subalgebra or P has not definite-parity.
    
    Note:
        The sup norm of P should be bounded by :math:`1 - \eta < 1`.
        The time required by the algorithm to compute the phase factors will scale with :math:`1/\eta`.
        
        The support_start of P will be ignored. In order to obtain phase factors for Laurent YQSP, first convert the polynomial into analytic form."""
    if not is_definite_parity(P):
        raise ValueError("Laurent polynomial is not of definite parity.")
    
    return yqsp_solve(laurent_to_analytic(P), mode=mode)

def chebqsp_solve(T: list[generic_complex] | ChebyshevTExpansion) -> ChebyshevQSPPhaseFactors:
    """Returns the set of phase factors for a Chebyshev QSP protocol implementing the polynomial `P(x)` (as the real part of the top-left polynomial, see Theorem 9 of arXiv:2105.02859).

    The target polynomial will be `P(x) = c[0] + c[1] T_1(x) + c[2] T_2(x) + ... + c[n] T_n(x)`, where `T_k` are the Chebyshev polynomials of the first kind.
    
    Args:
        c: the list of coefficients of `P(x)` in the Chebyshev basis, or a ChebyshevTExpansion object.
        
    Raises:
        ValueError: If the target polynomial does not have definite parity or is not real."""
    if isinstance(T, list):
        T = ChebyshevTExpansion(T)

    if not T.is_real():
        raise ValueError("Only real polynomials are supported.")

    xqsp = xqsp_solve_laurent(T.to_laurent())
    return ChebyshevQSPPhaseFactors(xqsp.phi)

def chebqsp_approximate(f, deg: int) -> ChebyshevQSPPhaseFactors:
    r"""Approximate the given callable object `f` (which takes :math:`x \in [-1, 1]` and returns a complex number)
    and returns the Chebyshev QSP phase factors implementing an approximating polynomial of degree `deg`
    (as the real part of the top-left polynomial, see Theorem 9 of arXiv:2105.02859).
    
    Note: The parity of `deg` should coincide with the parity of f, otherwise the Chebyshev approximator might give numerical errors."""
    return chebqsp_solve(chebyshev_approximate(f, deg))

def qsvt_solve(T: list[generic_complex] | ChebyshevTExpansion) -> QSVTPhaseFactors:
    """Returns the set of phase factors for a reflection QSP/QSVT protocol implementing the polynomial `P(x)` (as the real part of the top-left polynomial, see Theorem 9 of arXiv:2105.02859).

    The target polynomial will be `P(x) = c[0] + c[1] T_1(x) + c[2] T_2(x) + ... + c[n] T_n(x)`, where `T_k` are the Chebyshev polynomials of the first kind.
    
    Args:
        c: the list of coefficients of `P(x)` in the Chebyshev basis, or a ChebyshevTExpansion object.
    
    Raises:
        ValueError: If the target polynomial does not have definite parity or is not real."""
    return QSVTPhaseFactors.from_chebqsp(chebqsp_solve(T))

def qsvt_approximate(f, deg: int) -> QSVTPhaseFactors:
    r"""Approximate the given callable object `f` (which takes :math:`x \in [-1, 1]` and returns a complex number)
    and returns the reflection QSP (a.k.a. QSVT) phase factors implementing an approximating polynomial of degree `deg`
    (as the real part of the top-left polynomial, see Theorem 9 of arXiv:2105.02859).
    
    Note: The parity of `deg` should coincide with the parity of f, otherwise the Chebyshev approximator might give numerical errors."""
    return qsvt_solve(chebyshev_approximate(f, deg))



import unittest

import nlft_qsp.numerics as bd

from nlft_qsp.poly import ChebyshevTExpansion, Polynomial
from nlft_qsp.rand import random_sequence
from nlft_qsp.nlft import NonLinearFourierSequence


class PolynomialTestCase(unittest.TestCase):

    def test_get_set(self):
        p = Polynomial([1, 2, 3, 4], support_start=-3)

        self.assertEqual(p[-3], 1)
        self.assertEqual(p[-4], 0)
        self.assertEqual(p[-1], 3)
        self.assertEqual(p.effective_degree(), 3)

        p[-2] = 7
        self.assertEqual(p[-2], 7)
        self.assertEqual(p[-1], 3)
        self.assertEqual(p.effective_degree(), 3)

        p[1] = 9
        p[5] = 10
        self.assertEqual(p[1], 9)
        self.assertEqual(p[5], 10)
        self.assertEqual(p[3], 0)
        self.assertEqual(p.effective_degree(), 8)

        p[-5] = 20
        self.assertEqual(p[-5], 20)
        self.assertEqual(p[5], 10)
        self.assertEqual(p[-4], 0)
        self.assertEqual(p.effective_degree(), 10)

    def test_conjugate(self):
        p = Polynomial(random_sequence(10, 6), support_start=2)

        q = p.conjugate()

        for k in range(10):
            self.assertEqual(q[k], bd.conj(p[-k]))

    def test_truncate(self):
        p = Polynomial(range(20), support_start=-10)

        q = p.truncate(-5, 5)

        for k in p.support():
            if k not in q.support():
                self.assertEqual(q[k], 0)
            else:
                self.assertEqual(q[k], p[k])

    def test_add(self):
        p = bd.make_complex(1) + Polynomial([1, 2, 3])
        self.assertEqual(p.coeffs, [2, 2, 3])
        self.assertEqual(p.support_start, 0)

        p = Polynomial([1, 2, 3], support_start=2) - 4
        self.assertEqual(p.coeffs, [-4, 0, 1, 2, 3])
        self.assertEqual(p.support_start, 0)

        p = Polynomial([1, 2, 3], support_start=-3) + 7
        self.assertEqual(p.coeffs, [1, 2, 3, 7])
        self.assertEqual(p.support_start, -3)


        p = Polynomial([1, 2, 3], support_start=-3)
        q = p - Polynomial([4, 5, 6], support_start=-1)
        self.assertEqual(q.coeffs, [1, 2, -1, -5, -6])
        self.assertEqual(q.support_start, -3)

        q = p + Polynomial([4, 5, 6], support_start=-3)
        self.assertEqual(q.coeffs, [5, 7, 9])
        self.assertEqual(q.support_start, -3)

        q = p + Polynomial([4, 5, 6], support_start=-4)
        self.assertEqual(q.coeffs, [4, 6, 8, 3])
        self.assertEqual(q.support_start, -4)

    @bd.workdps(30)
    def test_mul(self):
        p = 3 * Polynomial([1, 2, 3], support_start=-10)
        self.assertEqual(p.coeffs, [3, 6, 9])
        self.assertEqual(p.support_start, -10)

        p = Polynomial([1, 2, 3], support_start=-1)
        q = p * Polynomial([5, 6, 7, 8], support_start=-5)

        for a, b in zip(q.coeffs, [5, 16, 34, 40, 37, 24]):
            self.assertAlmostEqual(a, b, delta=1e-25)
        self.assertEqual(q.support_start, -6)


    @bd.workdps(30)
    def test_call(self):
        p = Polynomial([3, 2, 1], support_start=0)
        q = Polynomial([5, -1, 0, 2], support_start=-3)

        self.assertAlmostEqual(p(1+1j), 5+4j, delta=10 * bd.machine_threshold())
        self.assertAlmostEqual(q(1+1j), 0.75-0.75j, delta=10 * bd.machine_threshold())

    @bd.workdps(30)
    def test_eval_at_roots_of_unity(self):
        seq = random_sequence(160000, 15)

        p = Polynomial(seq, support_start=0)
        q = Polynomial(seq, support_start=2)

        ep = p.eval_at_roots_of_unity(16)
        eq = q.eval_at_roots_of_unity(16)

        cep = [p(z) for z in bd.unitroots(16)]
        ceq = [q(z) for z in bd.unitroots(16)]
        for p1, p2 in zip(ep, cep):
            self.assertAlmostEqual(p1, p2, delta=10 * bd.machine_threshold())

        for z, q1, q2 in zip(bd.unitroots(16), eq, ceq):
            self.assertAlmostEqual(q1, q2, delta=10 * bd.machine_threshold())

        for z, a, b in zip(bd.unitroots(16), ep, eq):
            self.assertAlmostEqual(a * (z ** 2), b, delta=10 * bd.machine_threshold())


        ep = p.eval_at_roots_of_unity(8)
        eq = q.eval_at_roots_of_unity(8)

        cep = [p(z) for z in bd.unitroots(8)]
        ceq = [q(z) for z in bd.unitroots(8)]
        for p1, p2 in zip(ep, cep):
            self.assertAlmostEqual(p1, p2, delta=10 * bd.machine_threshold())

        for q1, q2 in zip(eq, ceq):
            self.assertAlmostEqual(q1, q2, delta=10 * bd.machine_threshold())

        for z, a, b in zip(bd.unitroots(8), ep, eq):
            self.assertAlmostEqual(a * (z ** 2), b, delta=10 * bd.machine_threshold())

    def test_schwarz_transform(self):
        p = Polynomial([])
        q = p.schwarz_transform()
        self.assertEqual(q.coeffs, [])

        p = Polynomial([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], support_start=-5)
        q = p.schwarz_transform()

        for k in p.support():
            if k < 0:
                self.assertEqual(q[k], 2*p[k])
            elif k == 0:
                self.assertEqual(q[k], p[k])
            else:
                self.assertEqual(q[k], 0)

    def test_chebyshev_expansion_to_laurent(self):
        T = ChebyshevTExpansion(random_sequence(10, 10))
        P = T.to_laurent()

        for k in range(32):
            theta = (k * bd.pi()) / 32
            z = bd.exp(1j * theta)
            x = bd.cos(theta)

            self.assertAlmostEqual(P(z), T(x), delta=bd.machine_threshold())

    def test_laurent_to_chebyshev_expansion(self):
        seq = random_sequence(10, 10)
        P = Polynomial(list(reversed(seq)) + seq[1:], support_start=-9)
        T = ChebyshevTExpansion(P)

        for k in range(32):
            theta = (k * bd.pi()) / 32
            z = bd.exp(1j * theta)
            x = bd.cos(theta)

            self.assertAlmostEqual(P(z), T(x), delta=bd.machine_threshold())


class NLFTTestCase(unittest.TestCase):

    @bd.workdps(10)
    def test_transform(self):
        nlft = NonLinearFourierSequence([])
        a, b = nlft.transform()
        self.assertEqual((a - 1).l2_squared_norm(), 0)
        self.assertEqual(b.l2_squared_norm(), 0)

        nlft = NonLinearFourierSequence([1, 2, 3, 4], support_start=0)
        a, b = nlft.transform()

        self.assertEqual(a.effective_degree(), 3)
        self.assertEqual(b.effective_degree(), 3)

        self.assertEqual(a.support_start, -3)
        self.assertEqual(b.support_start, 0)

        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=10 * bd.machine_threshold())
        
        for ak, ahk in zip(a.coeffs, [-0.0970143, 0.315296, -0.485071, 0.0242536]):
            self.assertAlmostEqual(ak, ahk, delta=1e-5)

        for bk, bhk in zip(b.coeffs, [0.0242536, -0.388057, -0.703353, 0.0970143]):
            self.assertAlmostEqual(bk, bhk, delta=1e-5)


        nlft = NonLinearFourierSequence([2, 3, 4], support_start=1)
        a, b = nlft.transform()

        self.assertEqual(a.effective_degree(), 2)
        self.assertEqual(b.effective_degree(), 2)

        self.assertEqual(a.support_start, -2)
        self.assertEqual(b.support_start, 1)

        self.assertAlmostEqual((a * a.conjugate() + b * b.conjugate() - 1).l2_norm(), 0, delta=10 * bd.machine_threshold())

        for ak, ahk in zip(a.coeffs, [-0.274398, -0.617395, 0.0342997]):
            self.assertAlmostEqual(ak, ahk, delta=1e-5)

        for bk, bhk in zip(b.coeffs, [0.0685994, -0.720294, 0.137199]):
            self.assertAlmostEqual(bk, bhk, delta=1e-5)

    def test_cheb2poly(self):
        P = Polynomial(random_sequence(10, 16))
        T = ChebyshevTExpansion.from_polynomial(P)

        s = 100000
        for l in range(s):
            x = 2*l/s - 1
            self.assertAlmostEqual(P(x), T(x), delta=bd.machine_threshold())

    def test_poly2cheb(self):
        T = ChebyshevTExpansion(random_sequence(10, 16))
        P = T.to_polynomial()

        s = 100000
        for l in range(s):
            x = 2*l/s - 1
            self.assertAlmostEqual(P(x), T(x), delta=bd.machine_threshold())


if __name__ == '__main__':
    unittest.main()